"""evohome-async - validate the v2 API token manager."""

from __future__ import annotations

import json
import logging
import uuid
from http import HTTPMethod, HTTPStatus
from typing import TYPE_CHECKING
from unittest.mock import AsyncMock, patch

import pytest
from aioresponses import aioresponses

from evohome_cli.auth import CredentialsManager
from evohomeasync2 import exceptions as exc
from tests.const import HEADERS_CRED_V2, URL_CRED_V2

from .test_v2_urls_cred import POST_CREDS  # HACK, should be in const.py, or a fixture

if TYPE_CHECKING:
    from pathlib import Path

    import aiohttp
    from freezegun.api import FrozenDateTimeFactory

    from evohome_cli.auth import CacheDataT


async def test_get_auth_token(
    client_session: aiohttp.ClientSession,
    credentials: tuple[str, str],
    cache_file: Path,
    freezer: FrozenDateTimeFactory,
) -> None:
    """Test .get_access_token() and .is_access_token_valid() methods."""

    def server_response() -> dict[str, int | str]:
        """Return the server response to a valid authorization request."""
        return {
            "access_token": str(uuid.uuid4()),
            "token_type": "bearer",
            "expires_in": 1800,
            "refresh_token": str(uuid.uuid4()),
        }

    # start with an empty cache
    token_manager = CredentialsManager(
        *credentials, client_session, cache_file=cache_file
    )

    #
    # have not yet called get_access_token (so not loaded cache either)
    assert token_manager.is_token_valid() is False

    with aioresponses() as rsp:
        response = {
            "error": "invalid_grant",
        }
        rsp.post(URL_CRED_V2, status=HTTPStatus.UNAUTHORIZED, payload=response)

        with pytest.raises(exc.AuthenticationFailedError):
            await token_manager.get_access_token()

        rsp.assert_called_once_with(POST_CREDS[0], POST_CREDS[1], **POST_CREDS[2])

    assert token_manager.is_token_valid() is False

    #
    # test HTTPStatus.OK
    payload = server_response()

    with aioresponses() as rsp:
        rsp.post(URL_CRED_V2, payload=payload)

        assert await token_manager.get_access_token() == payload["access_token"]

        rsp.assert_called_once_with(POST_CREDS[0], POST_CREDS[1], **POST_CREDS[2])

    assert token_manager.is_token_valid() is True

    #
    # check doesn't invoke the URL again, as session_id still valid
    freezer.tick(600)  # advance time by 5 minutes

    with patch("aiohttp.ClientSession.post", new_callable=AsyncMock) as mok:
        assert await token_manager.get_access_token() == payload["access_token"]

        mok.assert_not_called()

    assert token_manager.is_token_valid() is True

    #
    # check does invoke the URL, as access token now expired
    freezer.tick(1200)  # advance time by another 10 minutes, 15 total

    assert token_manager.is_token_valid() is False

    #
    # check does invoke the URL, as access token now expired
    data_token = {  # to assert POST was called with this data
        "grant_type": "refresh_token",
        "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",
        "refresh_token": payload["refresh_token"],
    }
    payload = server_response()

    with aioresponses() as rsp:
        rsp.post(URL_CRED_V2, payload=payload)

        assert await token_manager.get_access_token() == payload["access_token"]

        rsp.assert_called_once_with(
            URL_CRED_V2, HTTPMethod.POST, headers=HEADERS_CRED_V2, data=data_token
        )

    assert token_manager.is_token_valid() is True

    #
    # test _clear_access_token()
    token_manager._clear_access_token()

    assert token_manager.is_token_valid() is False


async def test_token_manager(
    cache_data_expired: CacheDataT,
    cache_data_valid: CacheDataT,
    cache_file: Path,
    client_session: aiohttp.ClientSession,
    credentials: tuple[str, str],
    caplog: pytest.LogCaptureFixture,
    freezer: FrozenDateTimeFactory,
    tmp_path_factory: pytest.TempPathFactory,
) -> None:
    """Test the .load_access_token() and .save_access_token() methods."""

    cache_file = tmp_path_factory.getbasetemp() / ".evo-cache.tst"

    #
    # TEST 1: load an invalid cache...
    with cache_file.open("w") as f:
        f.write(json.dumps(cache_data_expired, indent=4))

    token_manager = CredentialsManager(
        *credentials, client_session, cache_file=cache_file
    )

    # have not yet called get_access_token (so not loaded cache either)
    assert token_manager.is_session_valid() is False

    await token_manager.load_from_cache()
    assert token_manager.is_token_valid() is False

    #
    # TEST 2: load a valid token cache
    with cache_file.open("w") as f:
        f.write(json.dumps(cache_data_valid, indent=4))

    token_manager = CredentialsManager(
        *credentials, client_session, cache_file=cache_file
    )

    await token_manager.load_from_cache()
    assert token_manager.is_token_valid() is True

    access_token = await token_manager.get_access_token()

    #
    # TEST 3: some time has passed, but token is not expired
    freezer.tick(600)  # advance time by 5 minutes
    assert token_manager.is_token_valid() is True

    assert await token_manager.get_access_token() == access_token

    #
    # TEST 4: test save_access_token() method
    freezer.tick(1800)  # advance time by 15 minutes, 20 mins total
    assert token_manager.is_token_valid() is False

    with (
        patch(
            "evohomeasync2.auth.AbstractTokenManager._post_access_token_request",
            new_callable=AsyncMock,
        ) as req,
        patch(
            "evohome_cli.auth.CredentialsManager.save_access_token",
            new_callable=AsyncMock,
        ) as wrt,
    ):
        req.return_value = {
            "access_token": "new_access_token...",
            "expires_in": 1800,
            "refresh_token": "new_refresh_token...",
            "token_type": "bearer",
        }

        with caplog.at_level(logging.DEBUG):
            assert await token_manager.get_access_token() == "new_access_token..."

            # fmt: off
            assert caplog.records[0].message == ("Fetching access_token...")
            assert caplog.records[1].message == (" - authenticating with the refresh_token")
            assert caplog.records[2].message.startswith("POST https://tccna.resideo.com/Auth/OAuth/Token")
            assert caplog.records[3].message == (" - access_token = new_access_token...")
            assert caplog.records[4].message.startswith(" - access_token_expires = ")
            # fmt: on

            assert len(caplog.records) == 5  # noqa: PLR2004

        req.assert_called_once()
        wrt.assert_called_once()

    assert token_manager.is_token_valid() is True

    #
    # TEST 5: look for warning message
    freezer.tick(2400)  # advance time by another 20 minutes
    assert token_manager.is_token_valid() is False

    with (
        patch(
            "evohomeasync2.auth.AbstractTokenManager._post_access_token_request",
            new_callable=AsyncMock,
        ) as req,
        patch(
            "evohome_cli.auth.CredentialsManager.save_access_token",
            new_callable=AsyncMock,
        ) as wrt,
    ):
        req.return_value = {
            "access_token": "newer_access_token...",
            "expires_in": 1800,
            "refresh_token": "newer_refresh_token...",
            # "token_type": "bearer",  # should throw the warning we're looking for
        }

        caplog.clear()
        with caplog.at_level(logging.DEBUG):
            assert await token_manager.get_access_token() == "newer_access_token..."

            # "payload may be invalid: required key not provided @ data['token_type']"
            assert "payload may be invalid" in caplog.records[2].message

        req.assert_called_once()
        wrt.assert_called_once()

    assert token_manager.is_token_valid() is True



"""An async client for the v2 Resideo TCC API."""

from __future__ import annotations


class _EvohomeBaseError(Exception):
    """The base class for all exceptions."""

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class EvohomeError(_EvohomeBaseError):
    """The base class for all exceptions."""


# These occur whilst a RESTful API call is being made
class _ApiRequestFailedError(EvohomeError):
    """The API request failed for some reason (no/invalid/unexpected response)."""

    def __init__(self, message: str, status: int | None = None) -> None:
        super().__init__(message)
        self.status = status  # useful, available if via aiohttp.ClientResponseError


class ApiRequestFailedError(_ApiRequestFailedError):  # a base exception, API failed
    """The API request failed for some reason (no/invalid/unexpected response).

    Could be caused by any aiohttp.ClientError, for example: ConnectionError.  If the
    cause was a ClientResponseError, then the `status` attr will have an integer value.
    """


class ApiRateLimitExceededError(ApiRequestFailedError):
    """The API request failed because the vendor's API rate limit was exceeded."""


class AuthenticationFailedError(_ApiRequestFailedError):
    """Unable to authenticate the user credentials (unable to obtain an access token).

    The cause could be any ApiRequestFailedError, including RateLimitExceeded.
    """


class BadUserCredentialsError(AuthenticationFailedError):
    """Unable to authenticate the user credentials (unknown client_id or wrong secret).

    Reauthenticating will not help as the user credentials are proven to be invalid.
    """


class BadApiSchemaError(ApiRequestFailedError):  # a base exception, API data bad
    """The received/supplied JSON is not as expected (e.g. missing a required key)."""


class BadApiResponseError(BadApiSchemaError):
    """The received JSON is not as expected (e.g. missing a required key)."""


class BadApiRequestError(BadApiSchemaError):
    """The supplied parameter(s) are as expected (e.g. unknown/unsupported mode)."""


class BadScheduleUploadedError(BadApiRequestError):
    """The supplied schedule JSON is invalid / was not accepted by the vendor."""


# These occur without / after a RESTful API call (e.g. a necessary call was not made)
class _ConfigStatusError(EvohomeError):  # bad/missing data
    """The config/status JSON is missing or somehow invalid (has it been fetched?)."""


class ConfigError(_ConfigStatusError):  # a base exception, bad config JSON
    """The config JSON is missing or somehow invalid (e.g. InvalidSchemaError)."""


class InvalidConfigError(ConfigError):
    """The system config JSON is missing/invalid (has it been fetched?).

    This is likely because the user has not yet been authenticated (or authentication
    has failed).
    """


class NoSingleTcsError(ConfigError):
    """There is no default TCS (e.g. the user has more than one location)."""


class StatusError(_ConfigStatusError):  # a base exception, bad status/schedule JSON
    """The status JSON is missing or somehow invalid (e.g. BadApiResponseSchemaError)."""


class InvalidStatusError(StatusError):
    """The status JSON is missing/invalid (has it been fetched?).

    This is likely because the user has not yet called `Location.update()`.
    """


class InvalidScheduleError(StatusError):
    """The schedule JSON is missing/invalid (has it been fetched?).

    This is likely because the user has not yet called `Zone.get_schedule()`.
    """



"""Provides handling of TCC v2 locations.

The entity hierarchy is: Location -> Gateway -> TCS -> DHW | Zone.
"""

from __future__ import annotations

import logging
from datetime import datetime as dt, tzinfo
from typing import TYPE_CHECKING

from aiozoneinfo import async_get_time_zone

from _evohome.helpers import camel_to_snake, convert_naive_dtm_strs_to_aware
from _evohome.time_zone import EvoZoneInfo, iana_tz_from_windows_tz

from .const import (
    SZ_COUNTRY,
    SZ_GATEWAY_ID,
    SZ_GATEWAYS,
    SZ_LOCATION_ID,
    SZ_LOCATION_INFO,
    SZ_NAME,
    SZ_TIME_ZONE,
    SZ_TIME_ZONE_ID,
    SZ_USE_DAYLIGHT_SAVE_SWITCHING,
)
from .gateway import Gateway
from .schemas import factory_loc_status
from .schemas.const import EntityType
from .schemas.typedefs import EvoTimeZoneInfoT
from .zone import EntityBase

if TYPE_CHECKING:
    import voluptuous as vol

    from . import EvohomeClient
    from .schemas.typedefs import (
        EvoLocConfigEntryT,
        EvoLocConfigResponseT,
        EvoLocStatusResponseT,
        EvoTimeZoneInfoT,
    )


_LOGGER = logging.getLogger(__name__.rpartition(".")[0])


async def _create_tzinfo(
    time_zone_info: EvoTimeZoneInfoT, /, *, use_dst_switching: bool | None = False
) -> tzinfo:  # EvoZoneInfo | ZoneInfo:
    """Return a ZoneInfo object based on the time zone information.

    Note that `dst_enabled` (DST will be active at some times) is distinct from
    `dst_active` (DST is active now).
    """

    time_zone_id = time_zone_info[SZ_TIME_ZONE_ID]

    try:
        return await async_get_time_zone(iana_tz_from_windows_tz(time_zone_id))
    except (KeyError, ModuleNotFoundError):
        pass

    # DST support will be very limited, for example:
    # - unable to adjust current utcoffset when DST starts & stops
    # - unable to determine if a given datetime was/will be during DST

    msg = f"Unable to find IANA TZ identifier for '{time_zone_id}'"
    _LOGGER.warning(
        msg + "; DST support will be very limited" if use_dst_switching else ""
    )

    return EvoZoneInfo(
        time_zone_info=time_zone_info, use_dst_switching=use_dst_switching
    )


async def create_location(
    client: EvohomeClient,
    config: EvoLocConfigResponseT,
) -> Location:
    """Create a Location entity and return it.

    We use a constructor function to keep the async creation of a tzinfo object tightly
    coupled with the instantiation of its Location.
    """

    tzinfo = await _create_tzinfo(
        config[SZ_LOCATION_INFO][SZ_TIME_ZONE],
        use_dst_switching=config[SZ_LOCATION_INFO][SZ_USE_DAYLIGHT_SAVE_SWITCHING],
    )

    loc = Location(client, config, tzinfo=tzinfo)

    _LOGGER.debug(f"Instantiated {loc}")

    return loc


class Location(EntityBase):
    """Instance of an account's location."""

    SCH_STATUS: vol.Schema = factory_loc_status(camel_to_snake)
    _TYPE = EntityType.LOC

    def __init__(
        self,
        client: EvohomeClient,
        config: EvoLocConfigResponseT,
        /,
        *,
        tzinfo: tzinfo | None = None,
    ) -> None:
        super().__init__(
            config[SZ_LOCATION_INFO][SZ_LOCATION_ID],
            client.auth,
            client._logger,
        )

        self.client = client  # proxy for parent
        #

        # children
        self.gateways: list[Gateway] = []
        self.gateway_by_id: dict[str, Gateway] = {}

        self._config: EvoLocConfigEntryT = config[SZ_LOCATION_INFO]  # ?exclude TZ/DST

        self._tzinfo = tzinfo or EvoZoneInfo(
            time_zone_info=config[SZ_LOCATION_INFO][SZ_TIME_ZONE],
            use_dst_switching=config[SZ_LOCATION_INFO][SZ_USE_DAYLIGHT_SAVE_SWITCHING],
        )

        for gwy_entry in config[SZ_GATEWAYS]:
            gwy = Gateway(self, gwy_entry)

            self.gateways.append(gwy)
            self.gateway_by_id[gwy.id] = gwy

        self._status: EvoLocStatusResponseT | None = None

    def __str__(self) -> str:
        """Return a string representation of the entity."""
        return f"{self.__class__.__name__}(id='{self._id}', tzinfo='{self.tzinfo}')"

    @property
    def config(self) -> EvoLocConfigEntryT:
        """Return the latest config of the entity."""
        return self._config

    @property
    def status(self) -> EvoLocStatusResponseT:
        """Return the latest status of the entity."""
        return super().status  # type: ignore[return-value]

    # Config attrs...

    @property
    def country(self) -> str:
        # "Belgium"
        # "CzechRepublic"
        # "Netherlands"
        # "UnitedKingdom"

        return self._config[SZ_COUNTRY]

    @property
    def name(self) -> str:
        return self._config[SZ_NAME]

    async def _get_config(self) -> EvoLocConfigResponseT:
        """Get the latest config of the location and update its TZ/DST attrs.

        Usually called when DST starts/stops or the location's DST config changes (i.e.
        there is no _update_config() method). Returns the raw JSON of the latest config.
        """

        # it is assumed that *only* the location's TZ/DST info can change
        # so don't need ?includeTemperatureControlSystems=True

        config: EvoLocConfigResponseT = await self._auth.get(
            f"location/{self._id}/installationInfo"  # TODO: add schema
        )  # type: ignore[assignment]

        self._config = config[SZ_LOCATION_INFO]  # ?exclude TZ/DST

        # new TzInfo object, or update the existing one?
        self._tzinfo = await _create_tzinfo(
            config[SZ_LOCATION_INFO][SZ_TIME_ZONE],
            use_dst_switching=config[SZ_LOCATION_INFO][SZ_USE_DAYLIGHT_SAVE_SWITCHING],
        )
        # self._tzinfo._update(  # but, only for EvoZoneInfo...
        #     time_zone_info=config[SZ_LOCATION_INFO][SZ_TIME_ZONE],
        #     use_dst_switching=config[SZ_LOCATION_INFO][SZ_USE_DAYLIGHT_SAVE_SWITCHING],
        # )

        return config

    def _update_config(self, config: EvoLocConfigResponseT) -> None:
        """Update the LOC's config and cascade to its descendants."""
        # pass

    @property
    def tzinfo(self) -> tzinfo:
        """Return a tzinfo-compliant object for this Location."""
        return self._tzinfo

    def now(self) -> dt:  # always returns a TZ-aware dtm
        """Return the current local time as an aware datetime in this location's TZ."""
        return dt.now(self.client._tzinfo).astimezone(self.tzinfo)

    # Status (state) attrs & methods...

    async def update(
        self, *, _update_time_zone_info: bool = False
    ) -> EvoLocStatusResponseT:
        """Get the latest state of the location and update its status attrs.

        Will also update the status of its gateways, their TCSs, and their DHW/zones.
        Returns the raw JSON of the latest state.
        """

        if _update_time_zone_info:
            await self._get_config()

        status = await self._get_status()

        self._update_status(status)
        return status

    async def _get_status(self) -> EvoLocStatusResponseT:
        """Get the latest state of the location and update its status attr.

        Returns the raw JSON of the latest state.
        """

        status: EvoLocStatusResponseT = await self._auth.get(
            f"{self._TYPE}/{self.id}/status?includeTemperatureControlSystems=True",
            schema=self.SCH_STATUS,
        )  # type: ignore[assignment]

        self._update_status(status)
        return status

    def _update_status(self, status: EvoLocStatusResponseT) -> None:
        """Update the LOC's status and cascade to its descendants."""

        # convert all naive datetimes to TZ-aware datetimes (do when snake_casing?)
        status = convert_naive_dtm_strs_to_aware(status, self.tzinfo)

        # No ActiveFaults in location node of status

        # break the TypedDict into its parts (so, ignore[misc])...
        for gwy_status in status.pop(SZ_GATEWAYS):  # type: ignore[misc]
            if gwy := self.gateway_by_id.get(gwy_status[SZ_GATEWAY_ID]):
                gwy._update_status(gwy_status)

            else:
                self._logger.warning(
                    f"{self}: gateway_id='{gwy_status[SZ_GATEWAY_ID]} not known"
                    ", (has the location configuration changed?)"
                )

        self._status = status



"""A hacked aiohttp to provide a faked server."""

from __future__ import annotations

import asyncio
import json
from enum import EnumCheck, StrEnum, verify
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Self

if TYPE_CHECKING:
    from asyncio.streams import _ReaduntilBuffer
    from collections.abc import Generator
    from types import TracebackType

    from .const import _bodyT, _methodT, _statusT, _urlT
    from .vendor import FakedServer


_DEFAULT_LIMIT = 2**16  # 64 KiB


@verify(EnumCheck.UNIQUE)
class hdrs(StrEnum):  # noqa: N801  # is a la aiohttp
    METH_GET = "GET"
    METH_POST = "POST"
    METH_PUT = "PUT"


class StreamReader(asyncio.StreamReader):
    """A faked StreamReader."""

    _buffer = bytearray()

    def __init__(
        self, limit: int = _DEFAULT_LIMIT, loop: asyncio.AbstractEventLoop | None = None
    ) -> None:
        pass

    async def read(self, n: int = -1) -> bytes:
        """Read up to `n` bytes from the stream."""
        raise NotImplementedError

    async def readline(self) -> bytes:
        """Read chunk of data from the stream until newline (b'\n') is found."""
        return await self.readuntil(b"\n")

    async def readexactly(self, n: int) -> bytes:
        """Read exactly `n` bytes."""
        raise NotImplementedError

    async def readuntil(self, separator: _ReaduntilBuffer = b"\n") -> bytes:  # type: ignore[override]
        """Read data from the stream until ``separator`` is found."""
        raise NotImplementedError

    async def __aenter__(self, *args: Any, **kwargs: Any) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        pass


class ClientError(Exception):
    """A faked ClientError."""


class ClientResponseError(ClientError):
    """A faked ClientResponseError."""

    def __init__(
        self, msg: str, /, *, status: int | None = None, **kwargs: Any
    ) -> None:
        super().__init__(msg)
        self.status = status
        self.message = msg


class ContentTypeError(ClientResponseError):
    """ContentType found is not valid."""


class ClientTimeout:
    """A faked ClientTimeout."""

    def __init__(self, /, *, total: float | None = None, **kwargs: Any) -> None:
        self.total: float = total or 30


class ClientSession:
    """A faked ClientSession."""

    def __init__(
        self, /, *, timeout: ClientTimeout | None = None, **kwargs: Any
    ) -> None:
        self._timeout = timeout or ClientTimeout()

        # this is required, so no .get()
        self._faked_server: FakedServer = kwargs["faked_server"]

    def request(self, method: hdrs, url: str, **kwargs: Any) -> ClientResponse:
        """Perform HTTP request."""
        if method == hdrs.METH_GET:
            return self.get(url, **kwargs)
        if method == hdrs.METH_PUT:
            return self.put(url, **kwargs)
        if method == hdrs.METH_POST:
            return self.post(url, **kwargs)
        raise NotImplementedError

    def get(self, url: str, **kwargs: Any) -> ClientResponse:
        assert not {k: v for k, v in kwargs.items() if k != "headers" and v is not None}
        return ClientResponse(hdrs.METH_GET, url, session=self)  # type: ignore[arg-type]

    def put(
        self, url: str, /, *, data: Any = None, json: Any = None, **kwargs: Any
    ) -> ClientResponse:
        assert not {k: v for k, v in kwargs.items() if k != "headers"}
        return ClientResponse(hdrs.METH_PUT, url, data=data or json, session=self)  # type: ignore[arg-type]

    def post(self, url: str, /, *, data: Any = None, **kwargs: Any) -> ClientResponse:
        assert not {k: v for k, v in kwargs.items() if k != "headers"}
        return ClientResponse(hdrs.METH_POST, url, data=data, session=self)  # type: ignore[arg-type]

    async def close(self) -> None:
        pass


class ClientResponse:
    """A faked ClientResponse."""

    charset: str = "utf-8"

    def __init__(
        self,
        method: _methodT,
        url: _urlT,
        /,
        *,
        data: str | None = None,
        json: dict[str, Any] | None = None,
        session: ClientSession | None = None,
        **kwargs: Any,
    ) -> None:
        self.method = method
        self.url = url
        self.session = session

        assert self.session is not None  # mypy
        self._faked_server = self.session._faked_server

        self.status: _statusT = None  # type: ignore[assignment]
        self._body: _bodyT | None = None

        # self.content = StreamReader(
        #     self._faked_server.request(method, url, data=data or json)
        # )
        self._body = self._faked_server.request(method, url, data=data or json)
        self.status = self._faked_server.status  # type: ignore[assignment]

    def raise_for_status(self) -> None:
        if self.status >= HTTPStatus.MULTIPLE_CHOICES:  # 300
            raise ClientResponseError(
                f"{self.method} {self.url}: {self.status}", status=self.status
            )

    @property
    def content_length(self) -> int | None:
        if self._body is None:
            return None
        if self.content_type == "text/plain":
            return len(self._body)
        return len(str(self._body))

    @property
    def content_type(self) -> str | None:
        """Return the Content-Type header of the response."""

        # if isinstance(self._body, bytes):
        #     return "application/octet-stream"
        if isinstance(self._body, (dict | list)):
            return "application/json"
        if not isinstance(self._body, str):
            raise NotImplementedError
        if self._body.strip().startswith("<html>"):
            return "text/html"
        return "text/plain"

    async def text(self, /, **kwargs: Any) -> str:  # assumes is JSON or plaintext
        """Return the response body as text."""
        if self.content_type == "application/json":
            return json.dumps(self._body, indent=4)
        if self.content_type in ("text/html", "text/plain"):
            return str(self._body)
        raise NotImplementedError

    async def json(
        self, /, **kwargs: Any
    ) -> dict[str, Any] | list[Any]:  # assumes is JSON or plaintext
        """Return the response body as json (a dict)."""
        if self.content_type == "application/json":
            return self._body  # type: ignore[return-value]
        assert not isinstance(self._body, (dict | list))  # mypy
        return json.loads(self._body)  # type: ignore[arg-type,no-any-return]

    async def read(self) -> bytes:
        """Read response payload."""
        if self._body is None:
            raise NotImplementedError
        return b""

    async def __aenter__(self, *args: Any, **kwargs: Any) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        pass

    def __await__(self) -> Generator[Any, Any, Self]:
        """Make this class awaitable."""
        return self._await_impl().__await__()

    async def _await_impl(self) -> Self:
        """Return the actual result."""
        return self

    async def wait_for_close(self) -> None:
        """Wait for the response to close."""

    def release(self) -> None:
        """Release the response."""



"""Schema for vendor's TCC v2 API - for GET/PUT schedule of Zone/DHW.

The convention for JSON keys is camelCase, but the API appears to be case-insensitive.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Final, TypedDict

import voluptuous as vol

from _evohome.helpers import noop

from .const import (
    S2_COOL_SETPOINT,
    S2_DAILY_SCHEDULES,
    S2_DAY_OF_WEEK,
    S2_DHW_STATE,
    S2_HEAT_SETPOINT,
    S2_SWITCHPOINTS,
    S2_TIME_OF_DAY,
    DayOfWeek,
    DhwState,
)

if TYPE_CHECKING:
    from collections.abc import Callable


#######################################################################################
# GET/PUT DHW / Zone Schedules...
#


class TccDhwSwitchpointT(TypedDict):
    dhwState: DhwState  # Off, On
    timeOfDay: str


class TccDhwDayOfWeekT(TypedDict):
    dayOfWeek: str
    switchpoints: list[TccDhwSwitchpointT]


class TccDhwDailySchedulesT(TypedDict):
    dailySchedules: list[TccDhwDayOfWeekT]


class TccZonSwitchpointT(TypedDict):
    heatSetpoint: float
    timeOfDay: str


class TccZonDayOfWeekT(TypedDict):
    dayOfWeek: str
    switchpoints: list[TccZonSwitchpointT]


class TccZonDailySchedulesT(TypedDict):
    dailySchedules: list[TccZonDayOfWeekT]


#
# These are returned from vendor's API (GET)...
def factory_dhw_schedule(fnc: Callable[[str], str] = noop) -> vol.Schema:
    """Factory for the DHW schedule schema."""

    SCH_GET_SWITCHPOINT_DHW: Final = vol.Schema(  # TODO: checkme
        {
            vol.Required(fnc(S2_DHW_STATE)): vol.In(DhwState),
            vol.Required(fnc(S2_TIME_OF_DAY)): vol.Datetime(format="%H:%M:00"),
        },
        extra=vol.PREVENT_EXTRA,
    )

    SCH_GET_DAY_OF_WEEK_DHW: Final = vol.Schema(
        {
            vol.Required(fnc(S2_DAY_OF_WEEK)): vol.In(DayOfWeek),
            vol.Required(fnc(S2_SWITCHPOINTS)): [SCH_GET_SWITCHPOINT_DHW],
        },
        extra=vol.PREVENT_EXTRA,
    )

    return vol.Schema(
        {
            vol.Required(fnc(S2_DAILY_SCHEDULES)): [SCH_GET_DAY_OF_WEEK_DHW],
        },
        extra=vol.PREVENT_EXTRA,
    )


def factory_zon_schedule(fnc: Callable[[str], str] = noop) -> vol.Schema:
    """Factory for the zone schedule schema."""

    SCH_GET_SWITCHPOINT_ZONE: Final = vol.Schema(
        {
            vol.Optional(fnc(S2_COOL_SETPOINT)): float,  # an extrapolation
            vol.Required(fnc(S2_HEAT_SETPOINT)): vol.All(
                float, vol.Range(min=5, max=35)
            ),
            vol.Required(fnc(S2_TIME_OF_DAY)): vol.Datetime(format="%H:%M:00"),
        },
        extra=vol.PREVENT_EXTRA,
    )

    SCH_GET_DAY_OF_WEEK_ZONE: Final = vol.Schema(
        {
            # l.Required(fnc(S2_DAY_OF_WEEK)): vol.All(int, vol.Range(min=0, max=6)),  # 0 is Monday
            vol.Required(fnc(S2_DAY_OF_WEEK)): vol.In(DayOfWeek),
            vol.Required(fnc(S2_SWITCHPOINTS)): [SCH_GET_SWITCHPOINT_ZONE],
        },
        extra=vol.PREVENT_EXTRA,
    )

    return vol.Schema(
        {
            vol.Required(fnc(S2_DAILY_SCHEDULES)): [SCH_GET_DAY_OF_WEEK_ZONE],
        },
        extra=vol.PREVENT_EXTRA,
    )


#
# These are as to be provided to the vendor's API (PUT)...
# This is after modified by evohome-client (PUT), an evohome-client anachronism?
def _out_factory_put_schedule_dhw(fnc: Callable[[str], str] = noop) -> vol.Schema:
    """Factory for the zone schedule schema."""

    SCH_PUT_SWITCHPOINT_DHW: Final = vol.Schema(  # TODO: checkme
        {
            vol.Required(fnc(S2_DHW_STATE)): vol.In(DhwState),
            vol.Required(fnc(S2_TIME_OF_DAY)): vol.Datetime(format="%H:%M:00"),
        },
        extra=vol.PREVENT_EXTRA,
    )

    SCH_PUT_DAY_OF_WEEK_DHW: Final = vol.Schema(
        {
            vol.Required(fnc(S2_DAY_OF_WEEK)): vol.All(
                int, vol.Range(min=0, max=6)
            ),  # 0 is Monday
            vol.Required(fnc(S2_SWITCHPOINTS)): [SCH_PUT_SWITCHPOINT_DHW],
        },
        extra=vol.PREVENT_EXTRA,
    )

    return vol.Schema(
        {
            vol.Required(fnc(S2_DAILY_SCHEDULES)): [SCH_PUT_DAY_OF_WEEK_DHW],
        },
        extra=vol.PREVENT_EXTRA,
    )


def _out_factory_put_schedule_zone(fnc: Callable[[str], str] = noop) -> vol.Schema:
    """Factory for the zone schedule schema."""

    SCH_PUT_SWITCHPOINT_ZONE: Final = vol.Schema(
        {  # NOTE: S2_HEAT_SETPOINT is not .capitalized()
            #
            vol.Required(S2_HEAT_SETPOINT): vol.All(float, vol.Range(min=5, max=35)),
            vol.Required(fnc(S2_TIME_OF_DAY)): vol.Datetime(format="%H:%M:00"),
        },
        extra=vol.PREVENT_EXTRA,
    )

    SCH_PUT_DAY_OF_WEEK_ZONE: Final = vol.Schema(
        {
            vol.Required(fnc(S2_DAY_OF_WEEK)): vol.All(
                int, vol.Range(min=0, max=6)
            ),  # 0 is Monday
            vol.Required(fnc(S2_SWITCHPOINTS)): [SCH_PUT_SWITCHPOINT_ZONE],
        },
        extra=vol.PREVENT_EXTRA,
    )

    return vol.Schema(
        {
            vol.Required(fnc(S2_DAILY_SCHEDULES)): [SCH_PUT_DAY_OF_WEEK_ZONE],
        },
        extra=vol.PREVENT_EXTRA,
    )



"""Validate the handling of the v2 APIs (URLs) for Authorization.

This is used to:
  a) document the RESTful API that is provided by the vendor
  b) confirm the faked server (if any) is behaving as per a)

Testing is at HTTP request layer (e.g. GET/PUT).
Everything to/from the RESTful API is in camelCase (so those schemas are used).
"""

from __future__ import annotations

from datetime import timedelta as td
from http import HTTPMethod, HTTPStatus
from typing import TYPE_CHECKING, Any

import pytest

import evohomeasync2 as evo2
from evohomeasync2 import schemas
from evohomeasync2.const import API_STRFTIME
from evohomeasync2.schemas import SystemMode, ZoneMode
from tests.const import _DBG_USE_REAL_AIOHTTP

from .common import should_fail_v2, should_work_v2, skipif_auth_failed

if TYPE_CHECKING:
    from tests.conftest import EvohomeClientv2


#######################################################################################


async def _test_usr_account(evo: EvohomeClientv2) -> None:
    """Test /userAccount"""

    # STEP 1:
    url = "userAccount"
    _ = await should_work_v2(
        evo.auth, HTTPMethod.GET, url, schema=schemas.TCC_GET_USR_ACCOUNT
    )
    # {
    #     'userId': '2263181',
    #     'username': 'nobody@nowhere.com',
    # ...
    #     'country': 'UnitedKingdom',
    #     'language': 'enGB'
    # }

    # STEP 2:
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, url, status=HTTPStatus.METHOD_NOT_ALLOWED
    )
    # {'message': "The requested resource does not support http method 'PUT'."}

    # STEP 3:
    url = "userXxxxxxx"  # NOTE: is a general test, and not a test specific to this URL
    _ = await should_fail_v2(
        evo.auth,
        HTTPMethod.GET,
        url,
        status=HTTPStatus.NOT_FOUND,
        content_type="text/html",  # exception to usual content-type
    )
    # '<!DOCTYPE html PUBLIC ...


async def _test_user_locations(evo: EvohomeClientv2) -> None:
    """Test /location/installationInfo?userId={user_id}"""

    # TODO: can't use .update(); in any case, should use URLs only
    url = "userAccount"
    user_info: dict[str, Any] = await should_work_v2(
        evo.auth,
        HTTPMethod.GET,
        url,
        schema=None,  # schema not re-tested here
    )  # type: ignore[assignment]

    #
    url = f"location/installationInfo?userId={user_info['userId']}"
    _ = await should_work_v2(
        evo.auth,
        HTTPMethod.GET,
        url,
        schema=None,  # schema not tested here
    )

    # url = f"location/{loc_id}/installationInfo"  # no TCS info
    # _ = await should_work_v2(
    #     evo.auth,
    #     HTTPMethod.GET,
    #     url,
    #     schema=None,  # schema not tested here
    # )

    #
    url += "&includeTemperatureControlSystems=True"
    _ = await should_work_v2(
        evo.auth, HTTPMethod.GET, url, schema=schemas.TCC_GET_USR_LOCATIONS
    )

    #
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, url, status=HTTPStatus.METHOD_NOT_ALLOWED
    )

    #
    url = "location/installationInfo"
    _ = await should_fail_v2(evo.auth, HTTPMethod.GET, url, status=HTTPStatus.NOT_FOUND)

    #
    url = "location/installationInfo?userId=1230000"
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.GET, url, status=HTTPStatus.UNAUTHORIZED
    )

    #
    url = "location/installationInfo?userId=xxxxxxx"
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.GET, url, status=HTTPStatus.BAD_REQUEST
    )

    #
    url = "location/installationInfo?xxxxXx=xxxxxxx"
    _ = await should_fail_v2(evo.auth, HTTPMethod.GET, url, status=HTTPStatus.NOT_FOUND)


async def _test_loc_status(evo: EvohomeClientv2) -> None:
    """Test /location/{loc.id}/status"""

    # TODO: remove .update() and use URLs only
    await evo.update(dont_update_status=True)

    loc = evo.locations[0]
    #

    url = f"location/{loc.id}/status"
    _ = await should_work_v2(
        evo.auth,
        HTTPMethod.GET,
        url,
        schema=None,  # schema not tested here
    )

    url += "?includeTemperatureControlSystems=True"
    _ = await should_work_v2(
        evo.auth, HTTPMethod.GET, url, schema=schemas.TCC_GET_LOC_STATUS
    )
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, url, status=HTTPStatus.METHOD_NOT_ALLOWED
    )

    url = f"location/{loc.id}"
    await should_fail_v2(
        evo.auth,
        HTTPMethod.GET,
        url,
        status=HTTPStatus.NOT_FOUND,
        content_type="text/html",  # exception to usual content-type
    )

    url = "location/1230000/status"
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.GET, url, status=HTTPStatus.UNAUTHORIZED
    )

    url = "location/xxxxxxx/status"
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.GET, url, status=HTTPStatus.BAD_REQUEST
    )

    url = f"location/{loc.id}/xxxxxxx"
    _ = await should_fail_v2(
        evo.auth,
        HTTPMethod.GET,
        url,
        status=HTTPStatus.NOT_FOUND,
        content_type="text/html",  # exception to usual content-type
    )


async def _test_tcs_status(evo: EvohomeClientv2) -> None:
    """Test GET /temperatureControlSystem/{tcs.id}/status

    Also tests PUT /temperatureControlSystem/{tcs.id}/mode
    """

    # TODO: remove .update() and use URLs only?
    await evo.update(dont_update_status=True)

    tcs: evo2.ControlSystem
    if not (tcs := evo.locations[0].gateways[0].systems[0]):
        pytest.skip("No available TCS found")

    #
    # STEP 0: Get/keep the current mode, so we can restore it later
    url = f"{tcs._TYPE}/{tcs.id}/status"

    old_status: dict[str, Any] = await should_work_v2(
        evo.auth, HTTPMethod.GET, url, schema=schemas.TCC_GET_TCS_STATUS
    )  # type: ignore[assignment]
    # {
    #      'systemId': '1234567',
    #      'zones': [...]
    #      'systemModeStatus': {...}
    #      'activeFaults': [],
    # }

    old_mode = {
        "systemMode": old_status["systemModeStatus"]["mode"],
        "permanent": old_status["systemModeStatus"]["isPermanent"],
    }
    if "timeUntil" in old_status["systemModeStatus"]:
        old_mode["timeUntil"] = old_status["systemModeStatus"]["timeUntil"]

    #
    # STEP 1: Change the mode, but with the wrong method
    url = f"{tcs._TYPE}/{tcs.id}/mode"

    _ = await should_fail_v2(
        evo.auth, HTTPMethod.GET, url, status=HTTPStatus.METHOD_NOT_ALLOWED
    )
    # {'message': "The requested resource does not support http method 'GET'."}

    #
    # STEP 2: Change the mode, but with missing request data (JSON)
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, url, json={}, status=HTTPStatus.BAD_REQUEST
    )
    # [  # NOTE: keys are (case-insensitive) PascalCase, not camelCase!!
    #     {'code': 'ParameterIsMissing', 'parameterName': 'SystemMode', 'message': 'Parameter is missing.'},
    #     {'code': 'ParameterIsMissing', 'parameterName': 'Permanent',  'message': 'Parameter is missing.'}
    # ]

    #
    # STEP 3: Change the mode, but with invalid request data (JSON)
    new_mode = {"systemMode": "xxxxx", "permanent": True}

    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, url, json=new_mode, status=HTTPStatus.BAD_REQUEST
    )
    # [{'code': 'InvalidInput', 'message': 'Error converting value "xxxxx" to...'}]

    #
    # STEP 4: Change the mode, but with semi-invalid request data (JSON)
    assert SystemMode.COOL not in tcs.modes
    new_mode = {"systemMode": "Cool", "permanent": True}

    _ = await should_fail_v2(
        evo.auth,
        HTTPMethod.PUT,
        url,
        json=new_mode,
        status=HTTPStatus.INTERNAL_SERVER_ERROR,
    )
    # {'message': 'An error has occurred.'}

    #
    # STEP 4: Change the mode, with valid request data (JSON) (permanent)
    assert SystemMode.AUTO in tcs.modes
    new_mode = {"systemMode": SystemMode.AUTO, "permanent": True}

    _ = await should_work_v2(evo.auth, HTTPMethod.PUT, url, json=new_mode)
    # {'id': '1588314363'}

    #
    # STEP 4: Change the mode, with valid request data (JSON) (temporary)
    assert SystemMode.AWAY in tcs.modes
    new_mode = {
        "systemMode": SystemMode.AWAY,
        "permanent": False,
        "timeUntil": (tcs.location.now() + td(hours=1)).strftime(API_STRFTIME),
    }

    _ = await should_work_v2(evo.auth, HTTPMethod.PUT, url, json=new_mode)
    # {'id': '1588315695'}

    #
    # STEP 5: Restore the original mode
    _ = await should_work_v2(evo.auth, HTTPMethod.PUT, url, json=old_mode)
    # {'id': '1588316616'}

    #
    # STEP 6: Change the mode, but without permission
    url = f"{tcs._TYPE}/1234567/mode"

    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, url, json=old_mode, status=HTTPStatus.UNAUTHORIZED
    )
    # [{
    #     'code': 'Unauthorized',
    #     'message': 'You are not allowed to perform this action.'
    # }]

    #
    # STEP 7: hange the mode, but with invalid URL
    url = f"{tcs._TYPE}/{tcs.id}/systemMode"
    _ = await should_fail_v2(
        evo.auth,
        HTTPMethod.PUT,
        url,
        json=old_mode,
        status=HTTPStatus.NOT_FOUND,
        content_type="text/html",  # exception to usual content-type
    )
    # '<!DOCTYPE html PUBLIC ...


async def _test_zone_status(evo: EvohomeClientv2) -> None:
    """Test /temperatureZone/{zone.id}/status

    Also tests /temperatureZone/{zone.id}/heatSetpoint
    """

    heat_setpoint: dict[str, float | str | None]  # TODO: TypedDict

    # TODO: remove .update() and use URLs only
    await evo.update()

    if not (zone := evo.locations[0].gateways[0].systems[0].zones[0]):
        pytest.skip("No available zones found")

    #
    url = f"{zone._TYPE}/{zone.id}/status"
    _ = await should_work_v2(
        evo.auth, HTTPMethod.GET, url, schema=schemas.TCC_GET_ZON_STATUS
    )
    # {
    #     'zoneId': '3432576',
    #     'temperatureStatus': {'temperature': 25.5, 'isAvailable': True},
    #     'activeFaults': [],
    #     'setpointStatus': {'targetHeatTemperature': 18.5, 'setpointMode': 'FollowSchedule'},
    #     'name': 'Main Room'
    # }

    #
    url = f"{zone._TYPE}/{zone.id}/heatSetpoint"

    heat_setpoint = {
        "setpointMode": ZoneMode.PERMANENT_OVERRIDE,
    }
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, url, json=heat_setpoint, status=HTTPStatus.BAD_REQUEST
    )
    # [{
    #     'code': 'HeatSetpointChangeTargetTemperatureNotSet',
    #     'message': 'Target temperature not specified when required'
    # }]

    heat_setpoint = {
        "setpointMode": ZoneMode.PERMANENT_OVERRIDE,
        "HeatSetpointValue": 19.5 if zone.temperature is None else zone.temperature,
        # "timeUntil": None,
    }
    _ = await should_work_v2(evo.auth, HTTPMethod.PUT, url, json=heat_setpoint)
    # {'id': '1588359054'}

    #
    heat_setpoint = {
        "setpointMode": ZoneMode.PERMANENT_OVERRIDE,
        "HeatSetpointValue": 99,
        "timeUntil": None,
    }
    _ = await should_work_v2(evo.auth, HTTPMethod.PUT, url, json=heat_setpoint)
    # {'id': '1588359054'}

    #
    heat_setpoint = {
        "setpointMode": ZoneMode.TEMPORARY_OVERRIDE,
        "HeatSetpointValue": 19.5,
    }
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, url, json=heat_setpoint, status=HTTPStatus.BAD_REQUEST
    )
    # [{
    #     'code': 'HeatSetpointChangeTimeUntilNotSet',
    #     'message': 'Time until not specified when required'
    # }]

    #
    heat_setpoint = {
        "setpointMode": "xxxxx",
        "HeatSetpointValue": 19.5,
    }
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, url, json=heat_setpoint, status=HTTPStatus.BAD_REQUEST
    )
    # [{'code': 'InvalidInput', 'message': 'Error converting value "xxxxxxx" to ..."}]

    #
    heat_setpoint = {
        "setpointMode": ZoneMode.FOLLOW_SCHEDULE,
        "HeatSetpointValue": 0.0,
        "timeUntil": None,
    }
    _ = await should_work_v2(evo.auth, HTTPMethod.PUT, url, json=heat_setpoint)
    # {'id': '1588365922'}


#######################################################################################


@skipif_auth_failed  # GET
async def test_usr_account(evohome_v2: EvohomeClientv2) -> None:
    """Test GET /userAccount"""

    await _test_usr_account(evohome_v2)


@skipif_auth_failed  # GET
async def test_usr_locations(evohome_v2: EvohomeClientv2) -> None:
    """Test GET /location/installationInfo"""

    await _test_user_locations(evohome_v2)


@skipif_auth_failed  # GET
async def test_loc_status(evohome_v2: EvohomeClientv2) -> None:
    """Test GET /location/{loc.id}/status"""

    await _test_loc_status(evohome_v2)


@skipif_auth_failed  # GET, PUT
async def test_tcs_status(evohome_v2: EvohomeClientv2) -> None:
    """Test GET /temperatureControlSystem/{tcs.id}/status

    Also tests PUT /temperatureControlSystem/{tcs.id}/mode
    """

    try:
        await _test_tcs_status(evohome_v2)

    except NotImplementedError:  # TODO: implement
        if _DBG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Mocked server API not implemented")


@skipif_auth_failed  # GET, PUT
async def test_zone_status(evohome_v2: EvohomeClientv2) -> None:
    """Test GET /temperatureZone/{zone.id}/status

    Also tests PUT /temperatureZone/{zone.id}/heatSetpoint
    """

    try:
        await _test_zone_status(evohome_v2)

    except NotImplementedError:  # TODO: implement
        if _DBG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Mocked server API not implemented")


# TODO: test_get_dhw_status( & test_put_dhw_state(



"""evohomeasync provides an async client for the v0 Resideo TCC API."""

from __future__ import annotations

from enum import EnumCheck, StrEnum, verify
from typing import TYPE_CHECKING, Any, Final, NewType, NotRequired, TypedDict, TypeVar

import voluptuous as vol

from _evohome.helpers import noop, redact

if TYPE_CHECKING:
    from collections.abc import Callable

_T = TypeVar("_T")

# Datetime format used by the vendor's API
API_STRFTIME: Final = "%Y-%m-%dT%H:%M:%SZ"


# TCC identifiers (Usr, Loc, Gwy, Sys, Zon|Dhw)
_DhwIdT = NewType("_DhwIdT", int)
_GatewayIdT = NewType("_GatewayIdT", int)
_LocationIdT = NewType("_LocationIdT", int)
_SystemIdT = NewType("_SystemIdT", int)  # domainId ??
_UserIdT = NewType("_UserIdT", int)
_ZoneIdT = NewType("_ZoneIdT", int)

# TCC other
_TaskIdT = NewType("_TaskIdT", str)  # TODO: int or str?


#
SZ_ALLOWED_MODES: Final = "allowedModes"
SZ_CHANGEABLE_VALUES: Final = "changeableValues"
SZ_CITY: Final = "city"
SZ_COUNTRY: Final = "country"

SZ_DEVICE_COUNT: Final = "deviceCount"
SZ_DEVICE_ID: Final = "deviceID"  # is ID, not Id
SZ_DEVICES: Final = "devices"
SZ_DOMAIN_ID: Final = "domainID"  # is ID, not Id

SZ_FIRSTNAME: Final = "firstname"
SZ_GATEWAY_ID: Final = "gatewayId"
SZ_INDOOR_TEMPERATURE: Final = "indoorTemperature"
SZ_IS_ACTIVATED: Final = "isActivated"

SZ_LASTNAME: Final = "lastname"
SZ_LATEST_EULA_ACCEPTED: Final = "latestEulaAccepted"
SZ_LOCATION_ID: Final = "locationID"  # is ID, not Id

SZ_NAME: Final = "name"

SZ_SESSION_ID: Final = "sessionId"
SZ_STATE: Final = "state"
SZ_STREET_ADDRESS: Final = "streetAddress"
SZ_TELEPHONE: Final = "telephone"

SZ_USER_ID: Final = "userID"  # is ID, not Id
SZ_USER_INFO: Final = "userInfo"
SZ_USER_LANGUAGE: Final = "userLanguage"
SZ_USERNAME: Final = "username"

SZ_ZIPCODE: Final = "zipcode"


#######################################################################################
# [
#     {
#         "locationID": l["locationID"],
#         "devices": [
#             {
#                 k: v
#                 for k, v in d.items()
#                 if k[-2:].lower() == "id" or k in ("instance", "name")
#             }
#             for d in l["devices"]
#         ],
#     }
#     for l in config
# ]


def factory_failure_response(fnc: Callable[[str], str] = noop) -> vol.Schema:
    """Factory for the code/message response schema."""

    entry = vol.Schema(
        {
            vol.Required("code"): str,
            vol.Required("message"): str,
        },
        extra=vol.PREVENT_EXTRA,
    )

    return vol.Schema(vol.All([entry], vol.Length(min=1)))


# GET api/accountInfo -> userAccountInfoResponse
def factory_user_account_info_response(
    fnc: Callable[[str], str] = noop,
) -> vol.Schema:
    """Schema for the response to GET api/accountInfo."""

    # username: an email address
    # country:  ISO 3166-1 alpha-2 format (e.g. GB)

    return vol.Schema(
        {
            vol.Required(fnc(SZ_USER_ID)): int,
            vol.Required(fnc(SZ_USERNAME)): vol.All(str, vol.Length(min=1), redact),
            vol.Required(fnc(SZ_FIRSTNAME)): vol.All(str, redact),
            vol.Required(fnc(SZ_LASTNAME)): vol.All(str, redact),
            vol.Required(fnc(SZ_STREET_ADDRESS)): vol.All(str, redact),
            vol.Required(fnc(SZ_CITY)): vol.All(str, redact),
            # l.Optional(fnc(SZ_STATE)): str,  # missing?
            vol.Required(fnc(SZ_ZIPCODE)): vol.All(str, redact),
            vol.Required(fnc(SZ_COUNTRY)): vol.All(str, vol.Length(min=2)),
            vol.Required(fnc(SZ_TELEPHONE)): vol.All(str, redact),
            vol.Required(fnc(SZ_USER_LANGUAGE)): str,
        },
        extra=vol.ALLOW_EXTRA,
    )


# POST api/session -> sessionResponse
def factory_session_response(
    fnc: Callable[[str], str] = noop,
) -> vol.Schema:
    """Schema for the response to POST api/session."""

    # securityQuestionX: usu. "notUsed", a sentinel value
    SCH_SECURITY_QUESTION = vol.Any("notUsed", vol.All(str, redact))

    SCH_USER_ACCOUNT_RESPONSE = factory_user_account_info_response(fnc).extend(
        {
            vol.Required(fnc(SZ_IS_ACTIVATED)): bool,
            vol.Optional(fnc(SZ_DEVICE_COUNT)): int,
            vol.Optional(fnc("tenantID")): int,
            vol.Optional(fnc("securityQuestion1")): SCH_SECURITY_QUESTION,
            vol.Optional(fnc("securityQuestion2")): SCH_SECURITY_QUESTION,
            vol.Optional(fnc("securityQuestion3")): SCH_SECURITY_QUESTION,
            vol.Required(fnc(SZ_LATEST_EULA_ACCEPTED)): bool,
        },
        extra=vol.ALLOW_EXTRA,
    )

    return vol.Schema(
        {
            vol.Required(fnc(SZ_SESSION_ID)): vol.All(str, redact),
            vol.Required(fnc(SZ_USER_INFO)): SCH_USER_ACCOUNT_RESPONSE,
        },
        extra=vol.ALLOW_EXTRA,
    )


def _factory_location_response(
    fnc: Callable[[str], str] = noop,
) -> vol.Schema:
    """Factory for the user's location schema."""

    return vol.Schema(
        {
            vol.Required(fnc(SZ_LOCATION_ID)): int,  # is ID, not Id
            vol.Required(fnc(SZ_NAME)): vol.All(str, vol.Length(min=1)),
            vol.Required(fnc(SZ_STREET_ADDRESS)): str,
            vol.Required(fnc(SZ_CITY)): str,
            vol.Optional(fnc(SZ_STATE)): str,  # Optional
            vol.Required(fnc(SZ_COUNTRY)): vol.All(str, vol.Length(min=2)),  # GB
            vol.Required(fnc(SZ_ZIPCODE)): str,
            vol.Required(fnc("type")): vol.In(["Commercial", "Residential"]),
            vol.Required(fnc("hasStation")): bool,
            vol.Required(fnc(SZ_DEVICES)): [dict],  # TODO: [DeviceResponse]
            vol.Required(fnc("oneTouchButtons")): list,
            vol.Required(fnc("weather")): {str: object},  # WeatherResponse
            vol.Required(fnc("daylightSavingTimeEnabled")): bool,
            vol.Required(fnc("timeZone")): {str: object},  # TimeZoneResponse
            vol.Required(fnc("oneTouchActionsSuspended")): bool,
            vol.Required(fnc("isLocationOwner")): bool,
            vol.Required(fnc("locationOwnerID")): int,  # ID, not Id
            vol.Required(fnc("locationOwnerName")): str,
            vol.Required(fnc("locationOwnerUserName")): vol.All(str, vol.Length(min=1)),
            vol.Required(fnc("canSearchForContractors")): bool,
            vol.Optional(fnc("contractor")): {str: dict},  # ContractorResponse,Optional
        },
        extra=vol.ALLOW_EXTRA,
    )


# GET api/locations?userId={userId}&allData=True -> list[locationResponse]
def factory_location_response_list(
    fnc: Callable[[str], str] = noop,
) -> vol.Schema:
    """Schema for the response to GET api/locations?userId={userId}&allData=True."""

    return vol.Schema(
        vol.All([_factory_location_response(fnc)], vol.Length(min=0)),
        extra=vol.ALLOW_EXTRA,
    )


#######################################################################################


TCC_FAILURE_RESPONSE: Final = factory_failure_response()
TCC_GET_USR_INFO: Final = factory_user_account_info_response()
TCC_GET_USR_LOCS: Final = factory_location_response_list()
TCC_POST_USR_SESSION: Final = factory_session_response()


# schema keys (start with a lower case letter)
SZ_COOL_SETPOINT: Final = "coolSetpoint"
SZ_HEAT_SETPOINT: Final = "heatSetpoint"

SZ_ID: Final = "id"  # is id, not Id/ID
SZ_MAC_ID: Final = "macID"  # is ID, not Id
SZ_MODE: Final = "mode"
SZ_NEXT_TIME: Final = "NextTime"
SZ_QUICK_ACTION: Final = "QuickAction"
SZ_QUICK_ACTION_NEXT_TIME: Final = "QuickActionNextTime"
SZ_SETPOINT: Final = "setpoint"
SZ_SPECIAL_MODES: Final = "SpecialModes"
SZ_STATUS: Final = "status"
SZ_TEMP: Final = "temp"
SZ_THERMOSTAT: Final = "thermostat"
SZ_THERMOSTAT_MODEL_TYPE: Final = "thermostatModelType"
SZ_VALUE: Final = "value"

# schema values (start with an upper case letter)
SZ_AUTO: Final = "Auto"
SZ_AUTO_WITH_ECO: Final = "AutoWithEco"
SZ_AWAY: Final = "Away"
SZ_CUSTOM: Final = "Custom"
SZ_DAY_OFF: Final = "DayOff"
SZ_HEATING_OFF: Final = "HeatingOff"


@verify(EnumCheck.UNIQUE)
class SystemMode(StrEnum):
    AUTO = SZ_AUTO
    AUTO_WITH_ECO = SZ_AUTO_WITH_ECO
    AWAY = SZ_AWAY
    CUSTOM = SZ_CUSTOM
    DAY_OFF = SZ_DAY_OFF
    HEATING_OFF = SZ_HEATING_OFF


#
SZ_DHW_OFF: Final = "DHWOff"
SZ_DHW_ON: Final = "DHWOn"
#
SZ_DOMESTIC_HOT_WATER: Final = "DOMESTIC_HOT_WATER"
SZ_EMEA_ZONE: Final = "EMEA_ZONE"
#
SZ_HOLD: Final = "Hold"
SZ_SCHEDULED: Final = "Scheduled"
SZ_TEMPORARY: Final = "Temporary"
#
SZ_HEAT: Final = "Heat"
SZ_OFF: Final = "Off"


#######################################################################################
# These the responses via the vendor's API; they have camelCase keys...


class TccFailureResponseT(TypedDict):
    """Typed dict for code/message responses from the vendor servers."""

    code: str
    message: str


class TccSessionResponseT(TypedDict):
    """POST api/session"""

    sessionId: str
    userInfo: TccUserAccountResponseT


class TccUserAccountInfoResponseT(TypedDict):  # NOTE: is not TccUserAccountResponseT
    """GET api/accountInfo"""

    userID: _UserIdT
    username: str  # email address
    firstname: str
    lastname: str
    streetAddress: str
    city: str
    # state: str  # missing?
    zipcode: str
    country: str  # GB
    telephone: str
    userLanguage: str


class TccUserAccountResponseT(TccUserAccountInfoResponseT):
    """GET api/userAccounts?userId={userId}"""

    isActivated: bool
    deviceCount: int  # NotRequired?
    tenantId: int  # NotRequired?
    securityQuestion1: str  # NotRequired?
    securityQuestion2: str  # NotRequired?
    securityQuestion3: str  # NotRequired?
    latestEulaAccepted: bool  # NotRequired?


class TccLocationResponseT(TypedDict):
    """GET api/locations?locationId={locationId}&allData=True"""

    locationID: _LocationIdT  # is ID, not Id
    name: str
    streetAddress: str
    city: str
    state: str
    country: str
    zipcode: str
    type: str  # LocationType: "Commercial" | "Residential"
    hasStation: bool
    devices: list[TccDeviceResponseT]
    weather: TccWeatherResponseT  # WeatherResponse
    daylightSavingTimeEnabled: bool
    timeZone: TccTimeZoneResponseT
    oneTouchActionsSuspended: bool
    isLocationOwner: bool
    locationOwnerID: int  # TODO: check is ID, not Id
    locationOwnerName: str
    locationOwnerUserName: str
    canSearchforcontractors: bool
    contractor: NotRequired[dict[str, Any]]  # ContractorResponse


class TccDeviceResponseT(TypedDict):
    deviceID: _DhwIdT | _ZoneIdT  # is ID, not Id
    gatewayId: _GatewayIdT
    thermostatModelType: str  # DOMESTIC_HOT_WATER or a zone
    deviceType: int
    name: str
    scheduleCapable: bool
    holdUntilCapable: bool
    thermostat: TccThermostatResponseT
    humidifier: dict[str, Any]  # HumidifierResponse
    dehumidifier: dict[str, Any]  # DehumidifierResponse
    fan: dict[str, Any]  # FanResponse
    schedule: dict[str, Any]  # ScheduleResponse
    alertSettings: dict[str, Any]  # AlertSettingsResponse
    isUpgrading: bool
    isAlive: bool
    thermostatVersion: str
    macID: str  # is ID, not Id
    locationID: _LocationIdT  # is ID, not Id
    domainID: int  # is ID, not Id
    instance: int
    serialNumber: str
    pcbNumber: str


class TccThermostatResponseT(TypedDict):
    units: str  # displayedUnits: Fahrenheit or Celsius
    indoorTemperature: float
    outdoorTemperature: float
    outdoorTemperatureAvailable: bool
    outdoorHumidity: float
    outdoorHumidityAvailable: bool
    indoorHumidity: float
    indoorTemperatureStatus: str  # Measured|NotAvailable|SensorError|SensorFault
    indoorHumidityStatus: str
    outdoorTemperatureStatus: str
    outdoorHumidityStatus: str
    isCommercial: bool
    allowedModes: list[str]  # ThermostatMode
    deadband: float
    minHeatSetpoint: float
    maxHeatSetpoint: float
    minCoolSetpoint: float
    maxCoolSetpoint: float
    coolRate: float
    heatRate: float
    is_pre_cool_capable: bool
    changeable_values: TccThermostatChangeableValues
    equipment_outputStatus: str  # Off | Heating | Cooling
    schedule_capable: bool
    vacationHoldChangeable: bool
    vacationHoldCancelable: bool
    scheduleHeatSp: float
    scheduleCoolSp: float
    serialNumber: str
    pcbNumber: str


class TccThermostatChangeableValues(TypedDict):
    """
    "changeableValues": {
        "mode": "Off",
        "heatSetpoint": {"value": 21.0, "status": "Scheduled"},
        "vacationHoldDays": 0,
    },
    """

    mode: str  # Off
    heatSetpoint: _TccSetpointDict
    vacationHoldDays: int


class _TccSetpointDict(TypedDict):
    value: float
    status: str  # Scheduled, Temporary, Hold, VacationHold


class TccWeatherResponseT(TypedDict):
    condition: str  # an enum
    temperature: float
    units: str  # Fahrenheit (precision 1.0) or Celsius (0.5)
    humidity: int
    phrase: str


class TccTimeZoneResponseT(TypedDict):
    id: str
    displayName: str
    offsetMinutes: int
    currentOffsetMinutes: int
    usingDaylightSavingTime: bool


#######################################################################################
# These are identical but have snake_case keys, not camelCase keys...


class EvoFailureDictT(TypedDict):
    """Typed dict for code/message responses from the vendor servers."""

    code: str
    message: str


class EvoSessionDictT(TypedDict):
    """POST api/session"""

    session_id: str
    user_info: EvoUserAccountDictT


class EvoUserAccountInfoDictT(TypedDict):  # NOTE: is not EvoUserAccountDictT
    """GET api/accountInfo"""

    user_id: _UserIdT
    username: str  # email address
    firstname: str
    lastname: str
    street_address: str
    city: str
    # state: str  # missing?
    zipcode: str
    country: str  # GB
    telephone: str
    user_language: str


class EvoUserAccountDictT(EvoUserAccountInfoDictT):  # NOT EvoUserAccountInfoT
    """GET api/userAccounts?userId={userId}"""

    is_activated: bool
    device_count: int  # NotRequired?
    tenant_id: int  # NotRequired?
    security_question_1: str  # NotRequired?
    security_question_2: str  # NotRequired?
    security_question_3: str  # NotRequired?
    latest_eula_accepted: bool  # NotRequired?


class EvoLocInfoDictT(TypedDict):  # c.f. TccLocationResponseT
    location_id: _LocationIdT
    name: str
    street_address: str
    city: str
    state: str
    country: str
    zipcode: str
    type: str  # LocationType: "Commercial" | "Residential"
    has_station: bool
    devices: list[EvoDevInfoDictT]
    weather: EvoWeatherDictT  # WeatherResponse
    daylight_saving_time_enabled: bool
    time_zone: EvoTimeZoneInfoDictT
    is_location_owner: bool
    locationOwner_id: int
    location_owner_name: str
    location_owner_user_name: str
    can_searchforcontractors: bool
    contractor: NotRequired[dict[str, Any]]  # ContractorResponse


class EvoGwyInfoDictT(TypedDict):  # c.f. TccDeviceResponseT
    gateway_id: _GatewayIdT
    device_type: int
    mac_id: str
    location_id: int
    serial_number: str
    pcb_number: str


# These keys are in the JSON, but not in the developer docs for the API
class EvoTcsInfoDictT(EvoLocInfoDictT):
    domain_id: int
    one_touch_actions_suspended: bool
    one_touch_buttons: list[str]
    thermostat_version: str


class EvoDevInfoDictT(EvoGwyInfoDictT):
    device_id: _DhwIdT | _ZoneIdT
    name: str
    thermostat_model_type: str  # DOMESTIC_HOT_WATER or a zone, e.g. EMEA_ZONE
    schedule_capable: bool
    hold_until_capable: bool
    thermostat: EvoThermostatInfoDictT
    humidifier: dict[str, Any]  # HumidifierResponse
    dehumidifier: dict[str, Any]  # DehumidifierResponse
    fan: dict[str, Any]  # FanResponse
    schedule: dict[str, Any]  # ScheduleResponse
    alert_settings: dict[str, Any]  # AlertSettingsResponse
    is_upgrading: bool
    is_alive: bool
    instance: int


class EvoThermostatInfoDictT(TypedDict):
    units: str  # displayedUnits: Fahrenheit or Celsius
    indoor_temperature: float
    outdoor_temperature: float
    outdoor_temperature_available: bool
    outdoor_humidity: float
    outdoor_Humidity_available: bool
    indoor_humidity: float
    indoor_temperature_status: str  # Measured|NotAvailable|SensorError|SensorFault
    indoor_humidity_status: str
    outdoor_temperature_status: str
    outdoor_humidity_status: str
    is_commercial: bool
    allowed_modes: list[str]  # ThermostatMode
    deadband: float
    min_heat_setpoint: float
    max_heat_setpoint: float
    min_cool_setpoint: float
    max_cool_setpoint: float
    cool_rate: float
    heat_rate: float
    is_pre_cool_capable: bool
    changeable_values: Any  # thermostatChangeableValues
    equipment_output_status: str  # Off | Heating | Cooling
    schedule_capable: bool
    vacation_hold_changeable: bool
    vacation_hold_cancelable: bool
    schedule_heat_sp: float
    schedule_cool_sp: float
    serial_number: str
    pcb_number: str


class EvoWeatherDictT(TypedDict):
    condition: str  # an enum
    temperature: float
    units: str  # Fahrenheit (precision 1.0) or Celsius (0.5)
    humidity: int
    phrase: str


class EvoTimeZoneInfoDictT(TypedDict):
    id: str
    display_name: str
    offset_minutes: int
    current_offset_minutes: int
    using_daylight_saving_time: bool



"""Constants for evohomeasync2."""

from typing import Final

from .schemas.const import API_STRFTIME as API_STRFTIME  # noqa: PLC0414

# HDR_STRFTIME: Final = "%Y-%m-%d %H:%M:%S"  # used by HTTP headers

# These keys are snake_case equivalents of schema strings
SZ_ACTIVE_FAULTS: Final = "active_faults"
SZ_ALLOWED_MODES: Final = "allowed_modes"
SZ_ALLOWED_SETPOINT_MODES: Final = "allowed_setpoint_modes"
SZ_ALLOWED_SYSTEM_MODES: Final = "allowed_system_modes"

SZ_CAN_BE_PERMANENT: Final = "can_be_permanent"
SZ_CAN_BE_TEMPORARY: Final = "can_be_temporary"
SZ_COUNTRY: Final = "country"
SZ_CURRENT_OFFSET_MINUTES: Final = "current_offset_minutes"

SZ_DAILY_SCHEDULES: Final = "daily_schedules"
SZ_DHW: Final = "dhw"
SZ_DHW_ID: Final = "dhw_id"
SZ_DHW_STATE: Final = "dhw_state"
SZ_DHW_STATE_CAPABILITIES_RESPONSE: Final = "dhw_state_capabilities_response"

SZ_FAULT_TYPE: Final = "fault_type"

SZ_GATEWAY_ID: Final = "gateway_id"
SZ_GATEWAY_INFO: Final = "gateway_info"
SZ_GATEWAYS: Final = "gateways"

SZ_HEAT_SETPOINT: Final = "heat_setpoint"

SZ_IS_AVAILABLE: Final = "is_available"
SZ_IS_PERMANENT: Final = "is_permanent"

SZ_LOCATION_ID: Final = "location_id"
SZ_LOCATION_INFO: Final = "location_info"
SZ_LOCATION_OWNER: Final = "location_owner"

SZ_MAC: Final = "mac"
SZ_MAX_HEAT_SETPOINT: Final = "max_heat_setpoint"
SZ_MIN_HEAT_SETPOINT: Final = "min_heat_setpoint"
SZ_MODE: Final = "mode"
SZ_MODEL_TYPE: Final = "model_type"

SZ_NAME: Final = "name"

SZ_OFFSET_MINUTES: Final = "offset_minutes"

SZ_SCHEDULE_CAPABILITIES: Final = "schedule_capabilities"
SZ_SCHEDULE_CAPABILITIES_RESPONSE: Final = "schedule_capabilities_response"
SZ_SETPOINT_CAPABILITIES: Final = "setpoint_capabilities"
SZ_SETPOINT_MODE: Final = "setpoint_mode"
SZ_SETPOINT_STATUS: Final = "setpoint_status"
SZ_SINCE: Final = "since"
SZ_STATE: Final = "state"
SZ_STATE_STATUS: Final = "state_status"
SZ_SYSTEM_ID: Final = "system_id"
SZ_SYSTEM_MODE: Final = "system_mode"
SZ_SYSTEM_MODE_STATUS: Final = "system_mode_status"

SZ_TARGET_HEAT_TEMPERATURE: Final = "target_heat_temperature"
SZ_TEMPERATURE: Final = "temperature"
SZ_TEMPERATURE_STATUS: Final = "temperature_status"
SZ_TEMPERATURE_CONTROL_SYSTEMS: Final = "temperature_control_systems"
SZ_TIME_OF_DAY: Final = "time_of_day"
SZ_TIME_UNTIL: Final = "time_until"
SZ_TIME_ZONE: Final = "time_zone"
SZ_TIME_ZONE_ID: Final = "time_zone_id"
SZ_TIMING_MODE: Final = "timing_mode"

SZ_UNTIL: Final = "until"
SZ_USE_DAYLIGHT_SAVE_SWITCHING: Final = "use_daylight_save_switching"
SZ_USER_ID: Final = "user_id"

SZ_ZONE_ID: Final = "zone_id"
SZ_ZONE_TYPE: Final = "zone_type"
SZ_ZONES: Final = "zones"


# These keys are not in the vendor's schema
SZ_ID: Final = "id"
SZ_PERMANENT: Final = "permanent"
SZ_SCHEDULE: Final = "schedule"
SZ_SETPOINT: Final = "setpoint"
SZ_THERMOSTAT: Final = "thermostat"



"""Tests for evohome-async - helper functions."""

from __future__ import annotations

import inspect
import json
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest
import yaml
from freezegun.api import FakeDatetime  # to check schedules, setpoints

if TYPE_CHECKING:
    import voluptuous as vol


def assert_schema(folder: Path, schema: vol.Schema, file_name: str) -> None:
    if not Path(folder).joinpath(file_name).is_file():
        pytest.skip(f"No {file_name} in: {folder.name}")

    with Path(folder).joinpath(file_name).open() as f:
        data: dict[str, Any] = json.load(f)  # is camelCase, as per vendor's schema

    _ = schema(data)


# yaml.add_representer(FakeDatetime, fake_datetime_representer)
def fake_datetime_representer(
    dumper: yaml.Dumper, data: FakeDatetime
) -> yaml.nodes.ScalarNode:
    """Represent a FakeDatetime object as a string."""
    return dumper.represent_scalar("tag:yaml.org,2002:str", data.isoformat())


yaml.add_representer(FakeDatetime, fake_datetime_representer)


def get_property_methods(obj: object) -> list[str]:
    """Return a list of property methods of an object."""
    return [
        name
        for name, value in inspect.getmembers(obj.__class__)
        if isinstance(value, property | cached_property)
    ]


_ATTRS_NOT_TO_SERIALIZE = ("zone_by_name",)  # is already zone_by_id


def serializable_attrs(obj: object) -> dict[str, str]:
    """Return a dictionary of serializable attributes of an object."""

    result = {}
    for k in get_property_methods(obj):  # + list(vars(obj).keys()):
        if not k.startswith("_") and k not in _ATTRS_NOT_TO_SERIALIZE:
            try:
                result[k] = yaml.dump(getattr(obj, k))
            except TypeError:  # not all attrs are serializable
                continue

    return result



"""evohome-async - validate the v2 API authentication flow."""

from __future__ import annotations

import logging
from datetime import UTC, datetime as dt, timedelta as td
from http import HTTPMethod, HTTPStatus
from typing import TYPE_CHECKING

import pytest
from aioresponses import aioresponses

from evohome_cli.auth import CredentialsManager
from evohomeasync2 import EvohomeClient, exceptions as exc
from tests.const import (
    HEADERS_BASE,
    HEADERS_CRED_V2,
    TEST_PASSWORD,
    TEST_USERNAME,
    URL_CRED_V2,
)

from .const import LOG_01, LOG_02, LOG_03, LOG_04, LOG_20, LOG_29, LOG_90, LOG_99

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator
    from pathlib import Path

    from evohome_cli.auth import CredentialsManager


_TEST_ACCESS_TOKEN = "-- access token --"  # noqa: S105
_TEST_REFRESH_TOKEN = "-- refresh token --"  # noqa: S105


POST_CREDS = (
    "https://tccna.resideo.com/Auth/OAuth/Token",
    HTTPMethod.POST,
    {
        "headers": HEADERS_CRED_V2,
        "data": {
            "grant_type": "password",
            "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",
            "Username": TEST_USERNAME,
            "Password": TEST_PASSWORD,
        },
    },
)

POST_REFRESH = (
    "https://tccna.resideo.com/Auth/OAuth/Token",
    HTTPMethod.POST,
    {
        "headers": HEADERS_CRED_V2,
        "data": {
            "grant_type": "refresh_token",
            "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",
            "refresh_token": _TEST_REFRESH_TOKEN,
        },
    },
)

GET_ACCOUNT = (
    "https://tccna.resideo.com/WebAPI/emea/api/v1/userAccount",
    HTTPMethod.GET,
    {
        "headers": HEADERS_BASE | {"Authorization": f"bearer {_TEST_ACCESS_TOKEN}"},
    },
)


@pytest.fixture(scope="module")
def cache_file(
    tmp_path_factory: pytest.TempPathFactory,
) -> Path:
    """Return the path to the token cache."""
    return tmp_path_factory.mktemp(__name__) / ".evo-cache.tst"


# NOTE: using fixture_folder will break these tests; we don't want evo.update() either
@pytest.fixture
async def evohome_v2(
    credentials_manager: CredentialsManager,
) -> AsyncGenerator[EvohomeClient]:
    """Yield a client with an vailla credentials manager."""

    evo = EvohomeClient(credentials_manager)

    try:
        yield evo
    finally:
        pass


async def test_bad1(  # bad credentials (client_id/secret)
    credentials: tuple[str, str],
    evohome_v2: EvohomeClient,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test authentication flow with bad credentials (client_id/secret)."""

    # pre-requisite data (no session_id)
    assert evohome_v2._token_manager.is_token_valid() is False

    # TEST 1: bad credentials (client_id/secret) -> HTTPStatus.UNAUTHORIZED
    with aioresponses() as rsp, caplog.at_level(logging.DEBUG):
        rsp.post(
            URL_CRED_V2,
            status=HTTPStatus.BAD_REQUEST,
            payload={"error": "invalid_grant"},
        )

        with pytest.raises(exc.BadUserCredentialsError) as err:
            await evohome_v2.update()

        assert err.value.status == HTTPStatus.BAD_REQUEST
        assert caplog.record_tuples == [LOG_01, LOG_04, LOG_90]
        assert len(rsp.requests) == 1

        # response 0: Unauthorized (bad credentials)
        rsp.assert_called_once_with(POST_CREDS[0], POST_CREDS[1], **POST_CREDS[2])

    assert evohome_v2._token_manager.is_token_valid() is False


async def test_bad2(  # bad access token
    credentials: tuple[str, str],
    evohome_v2: EvohomeClient,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test authentication flow with an invalid/expired access token."""

    # pre-requisite data (a valid access token that will nonetheless be rejected)
    evohome_v2._token_manager._access_token = _TEST_ACCESS_TOKEN
    evohome_v2._token_manager._access_token_expires = dt.now(tz=UTC) + td(minutes=15)
    evohome_v2._token_manager._refresh_token = _TEST_REFRESH_TOKEN

    assert evohome_v2._token_manager.is_token_valid() is True

    # TEST 9: bad access token -> HTTPStatus.UNAUTHORIZED
    with aioresponses() as rsp, caplog.at_level(logging.DEBUG):
        rsp.get(
            "https://tccna.resideo.com/WebAPI/emea/api/v1/userAccount",
            status=HTTPStatus.UNAUTHORIZED,
            payload=[{"code": "Unauthorized", "message": "Unauthorized"}],
        )

        with pytest.raises(exc.AuthenticationFailedError) as err:
            await evohome_v2.update()

        assert err.value.status is None  # Connection refused
        assert caplog.record_tuples == [LOG_20, LOG_01, LOG_02, LOG_99]
        assert len(rsp.requests) == 2  # noqa: PLR2004

        # response 0: Unauthorized (bad access token)
        rsp.assert_called_with(GET_ACCOUNT[0], GET_ACCOUNT[1], **GET_ACCOUNT[2])

        # response 1: Connection refused (as no response provided by us)
        rsp.assert_called_with(POST_REFRESH[0], POST_REFRESH[1], **POST_REFRESH[2])

    assert evohome_v2._token_manager.is_token_valid() is False


async def test_bad3(  # bad credentials (refresh token)
    credentials: tuple[str, str],
    evohome_v2: EvohomeClient,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test authentication flow with invalid/unknown credentials (refresh token)."""

    # pre-requisite data (an expired access token, with a bad refresh token)
    evohome_v2._token_manager._access_token = _TEST_ACCESS_TOKEN
    evohome_v2._token_manager._access_token_expires = dt.now(tz=UTC) - td(minutes=15)
    evohome_v2._token_manager._refresh_token = _TEST_REFRESH_TOKEN

    assert evohome_v2._token_manager.is_token_valid() is False

    # TEST 9: bad session id -> HTTPStatus.BAD_REQUEST
    with aioresponses() as rsp, caplog.at_level(logging.DEBUG):
        rsp.post(
            URL_CRED_V2,
            status=HTTPStatus.BAD_REQUEST,
            payload={"error": "invalid_grant"},
        )

        with pytest.raises(exc.AuthenticationFailedError) as err:
            await evohome_v2.update()

        assert err.value.status is None  # Connection refused
        assert caplog.record_tuples == [LOG_01, LOG_02, LOG_03, LOG_04, LOG_99]
        assert len(rsp.requests) == 1

        # response 0: invalid_grant (bad refresh token)
        rsp.assert_any_call(POST_REFRESH[0], POST_REFRESH[1], **POST_REFRESH[2])

        # response 1: Connection refused (as no response provided by us)
        rsp.assert_called_with(POST_CREDS[0], POST_CREDS[1], **POST_CREDS[2])

    assert evohome_v2._token_manager.is_token_valid() is False


async def test_good(  # good credentials
    credentials: tuple[str, str],
    evohome_v2: EvohomeClient,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test authentication flow (and authorization) with good credentials."""

    # pre-requisite data (no session_id)
    assert evohome_v2._token_manager.is_token_valid() is False

    #
    # TEST 1: good credentials (client_id/secret) -> HTTPStatus.OK
    with aioresponses() as rsp, caplog.at_level(logging.WARNING):
        rsp.post(
            URL_CRED_V2,
            status=HTTPStatus.OK,
            payload={
                "access_token": _TEST_ACCESS_TOKEN,
                "token_type": "bearer",
                "expires_in": 1799,
                "refresh_token": _TEST_REFRESH_TOKEN,
                # "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",  # optional
            },
        )

        with pytest.raises(exc.ApiRequestFailedError) as err:
            await evohome_v2.update()

        assert err.value.status is None  # Connection refused
        assert caplog.record_tuples == [LOG_29]
        assert len(rsp.requests) == 2  # noqa: PLR2004

        # response 0: Successful authentication
        rsp.assert_called_with(POST_CREDS[0], POST_CREDS[1], **POST_CREDS[2])

        # response 1: Connection refused (as no response provided by us)
        rsp.assert_called_with(GET_ACCOUNT[0], GET_ACCOUNT[1], **GET_ACCOUNT[2])

    assert evohome_v2._token_manager.is_token_valid() is True



"""Tests for evohome-async."""

from __future__ import annotations

import json
from datetime import UTC, datetime as dt, timedelta as td
from typing import TYPE_CHECKING

import pytest

from evohome_cli.auth import CredentialsManager
from evohomeasync import EvohomeClient as EvohomeClientv0
from evohomeasync2 import EvohomeClient as EvohomeClientv2

from .const import TEST_PASSWORD, TEST_USERNAME

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator
    from pathlib import Path

    import aiohttp

    from evohome_cli.auth import CacheDataT


@pytest.fixture  # @pytest_asyncio.fixture(scope="session", loop_scope="session")
async def client_session(
    use_real_aiohttp: bool,  # noqa: FBT001 (is a fixture)
) -> AsyncGenerator[aiohttp.ClientSession]:
    """Yield an aiohttp.ClientSession, which may be faked."""

    if use_real_aiohttp:
        import aiohttp  # noqa: PLC0415

        client_session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=30))

    else:
        from .tests_rf import faked_server as fake  # noqa: I001, PLC0415
        from .tests_rf.faked_server import aiohttp  # type: ignore[no-redef] # noqa: PLC0415

        client_session = aiohttp.ClientSession(faked_server=fake.FakedServer({}, {}))  # type: ignore[call-arg]

    try:
        yield client_session
    finally:
        await client_session.close()


@pytest.fixture(scope="session")
def credentials() -> tuple[str, str]:
    """Return a username and a password."""

    username: str = TEST_USERNAME
    password: str = TEST_PASSWORD

    return username, password


@pytest.fixture(scope="session")
def cache_data_valid(credentials: tuple[str, str]) -> CacheDataT:
    """Return the path to the token cache."""

    return {
        credentials[0]: {
            "access_token": {
                "access_token": "ncw...",
                "access_token_expires": (dt.now(tz=UTC) + td(minutes=15)).isoformat(),
                "refresh_token": "ryx...",
            },
            "session_id": {
                "session_id": "123...",
                "session_id_expires": (dt.now(tz=UTC) + td(minutes=15)).isoformat(),
            },
        },
        "username@gmail.com.xx": {
            "access_token": {
                "access_token": "ncw...",
                "access_token_expires": (dt.now(tz=UTC) + td(hours=1)).isoformat(),
                "refresh_token": "ryx...",
            },
            "session_id": {
                "session_id": "123...",
                "session_id_expires": (dt.now(tz=UTC) - td(hours=1)).isoformat(),
            },
        },
    }


@pytest.fixture(scope="session")
def cache_data_expired(credentials: tuple[str, str]) -> CacheDataT:
    """Return the path to the token cache."""

    return {
        credentials[0]: {
            "access_token": {
                "access_token": "ncw...",
                "access_token_expires": (dt.now(tz=UTC) - td(hours=1)).isoformat(),
                "refresh_token": "ryx...",
            },
            "session_id": {
                "session_id": "123...",
                "session_id_expires": (dt.now(tz=UTC) - td(hours=1)).isoformat(),
            },
        },
    }


@pytest.fixture(scope="session")
def cache_file(
    cache_data_valid: dict[str, int | str],
    tmp_path_factory: pytest.TempPathFactory,
    use_real_aiohttp: bool,  # noqa: FBT001 (is a fixture)
) -> Path:
    """Return the path to the token cache."""

    # don't pollute the cache of real tokens (from the vendor) with fake tokens
    if use_real_aiohttp:
        from evohome_cli.client import CACHE_FILE  # noqa: PLC0415

        return CACHE_FILE

    cache_file = tmp_path_factory.getbasetemp() / ".evo-cache.tst"

    with cache_file.open("w") as f:
        f.write(json.dumps(cache_data_valid, indent=4))

    return cache_file


@pytest.fixture  # @pytest_asyncio.fixture(scope="session", loop_scope="session")
async def credentials_manager(
    client_session: aiohttp.ClientSession,
    credentials: tuple[str, str],
    cache_file: Path,
) -> AsyncGenerator[CredentialsManager]:
    """Yield a credentials manager for access_token & session_id caching."""

    manager = CredentialsManager(*credentials, client_session, cache_file=cache_file)

    await manager.load_from_cache()

    try:
        yield manager
    finally:
        await manager.save_to_cache()  # for next run of tests


@pytest.fixture
async def evohome_v0(
    credentials_manager: CredentialsManager,
) -> AsyncGenerator[EvohomeClientv0]:
    """Yield an instance of a v0 EvohomeClient."""

    evo = EvohomeClientv0(credentials_manager)

    # await evo.update()

    try:
        yield evo
    finally:
        pass


@pytest.fixture
async def evohome_v2(
    credentials_manager: CredentialsManager,
) -> AsyncGenerator[EvohomeClientv2]:
    """Yield an instance of a v2 EvohomeClient."""

    evo = EvohomeClientv2(credentials_manager)

    # await evo.update()

    try:
        yield evo
    finally:
        pass



"""Validate the handling of the vendor's v2 APIs (URLs) for Task management.

This is used to:
  a) document the RESTful API that is provided by the vendor
  b) confirm the faked server (if any) is behaving as per a)

Testing is at HTTP request layer (e.g. GET/PUT).
Everything to/from the RESTful API is in camelCase (so those schemas are used).
"""

from __future__ import annotations

from datetime import timedelta as td
from http import HTTPMethod, HTTPStatus
from typing import TYPE_CHECKING

import pytest

import evohomeasync2 as evo2
from _evohome.helpers import camel_to_pascal
from evohomeasync2.const import API_STRFTIME
from evohomeasync2.schemas import DhwState, ZoneMode
from evohomeasync2.schemas.const import (
    S2_MODE,
    S2_STATE,
    S2_STATE_STATUS,
    S2_UNTIL,
    S2_UNTIL_TIME,
)
from tests.const import _DBG_USE_REAL_AIOHTTP

from .common import should_fail_v2, should_work_v2, skipif_auth_failed

if TYPE_CHECKING:
    from tests.conftest import EvohomeClientv2

#######################################################################################


# NOTE: a long test, but not all systems have DHW
async def _test_task_id_dhw(evo: EvohomeClientv2) -> None:
    """Test the task_id returned when using the vendor's RESTful APIs.

    This test can be used to prove that JSON keys are can be camelCase or PascalCase.
    """

    await evo.update(dont_update_status=True)

    dhw: evo2.HotWater | None = None

    for loc in evo.locations:
        for gwy in loc.gateways:
            for tcs in gwy.systems:
                if tcs.hotwater:
                    # if (dhw := tcs.hotwater) and dhw.temperatureStatus['isAvailable']:
                    dhw = tcs.hotwater
                    break

    if dhw is None:
        pytest.skip("No available DHW found")

    GET_URL = f"{dhw._TYPE}/{dhw.id}/status"
    PUT_URL = f"{dhw._TYPE}/{dhw.id}/state"

    #
    # PART 0: Get initial state...
    old_status = await should_work_v2(evo.auth, HTTPMethod.GET, GET_URL)
    assert isinstance(old_status, dict)  # mypy  TODO: use a SCHEMA
    # {
    #     'dhwId': '3933910',
    #     'temperatureStatus': {'isAvailable': False},
    #     'stateStatus': {'state': 'Off', 'mode': 'FollowSchedule'},
    #     'activeFaults': []
    # }  # HTTP 200
    # {
    #     'dhwId': '3933910',
    #     'temperatureStatus': {'temperature': 21.0, 'isAvailable': True},
    #     'stateStatus': {
    #         'state': 'On',
    #         'mode': 'TemporaryOverride',
    #         'until': '2023-10-30T18:40:00Z'
    #     },
    #     'activeFaults': []
    # }  # HTTP 200

    old_mode = {
        S2_MODE: old_status[S2_STATE_STATUS][S2_MODE],
        S2_STATE: old_status[S2_STATE_STATUS][S2_STATE],
        S2_UNTIL_TIME: old_status[S2_STATE_STATUS].get(S2_UNTIL),
    }  # NOTE: untilTime/until

    #
    # PART 1: Try the basic functionality...
    # new_mode = {S2_MODE: ZoneMode.PERMANENT_OVERRIDE, S2_STATE: DhwState.OFF, S2_UNTIL_TIME: None}
    new_mode = {
        S2_MODE: ZoneMode.TEMPORARY_OVERRIDE,
        S2_STATE: DhwState.ON,
        S2_UNTIL_TIME: (loc.now() + td(hours=1)).strftime(API_STRFTIME),
    }

    result = await should_work_v2(evo.auth, HTTPMethod.PUT, PUT_URL, json=new_mode)
    assert isinstance(result, dict | list)  # mypy
    # {'id': '840367013'}  # HTTP 201/Created

    task_id = result[0]["id"] if isinstance(result, list) else result["id"]
    url_tsk = f"commTasks?commTaskId={task_id}"

    assert int(task_id)

    status = await should_work_v2(evo.auth, HTTPMethod.GET, url_tsk)
    # {'commtaskId': '840367013', 'state': 'Created'}
    # {'commtaskId': '840367013', 'state': 'Succeeded'}

    assert isinstance(status, dict)  # mypy
    assert status["commtaskId"] == task_id
    assert status["state"] in ("Created", "Running", "Succeeded")

    # async with asyncio.timeout(30):
    #     _ = await wait_for_comm_task(evo.auth, task_id)

    #
    # PART 2A: Try different capitalisations of the JSON keys...
    new_mode = {
        S2_MODE: ZoneMode.TEMPORARY_OVERRIDE,
        S2_STATE: DhwState.ON,
        S2_UNTIL_TIME: (loc.now() + td(hours=1)).strftime(API_STRFTIME),
    }
    _ = await should_work_v2(
        evo.auth, HTTPMethod.PUT, PUT_URL, json=new_mode
    )  # HTTP 201

    # async with asyncio.timeout(30):
    #     _ = await wait_for_comm_task(evo.auth, task_id)

    status = await should_work_v2(evo.auth, HTTPMethod.GET, GET_URL)

    new_mode = {  # NOTE: different capitalisation, until time
        camel_to_pascal(S2_MODE): ZoneMode.TEMPORARY_OVERRIDE,
        camel_to_pascal(S2_STATE): DhwState.ON,
        camel_to_pascal(S2_UNTIL_TIME): (loc.now() + td(hours=2)).strftime(
            API_STRFTIME
        ),
    }
    _ = await should_work_v2(evo.auth, HTTPMethod.PUT, PUT_URL, json=new_mode)

    # async with asyncio.timeout(30):
    #     _ = await wait_for_comm_task(evo.auth, task_id)

    status = await should_work_v2(evo.auth, HTTPMethod.GET, GET_URL)

    #
    # PART 3: Restore the original mode
    _ = await should_work_v2(evo.auth, HTTPMethod.PUT, PUT_URL, json=old_mode)

    # async with asyncio.timeout(30):
    #    _ = await wait_for_comm_task(evo.auth, task_id)

    status = await should_work_v2(evo.auth, HTTPMethod.GET, GET_URL)

    # assert status # != old_status

    #
    # PART 4A: Try bad JSON...
    bad_mode = {
        S2_STATE: ZoneMode.TEMPORARY_OVERRIDE,
        S2_MODE: DhwState.OFF,
        S2_UNTIL_TIME: None,
    }
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, PUT_URL, json=bad_mode, status=HTTPStatus.BAD_REQUEST
    )

    # _ = [{
    #     "code": "InvalidInput", "message": """
    #         Error converting value 'TemporaryOverride'
    #         to type 'DomesticHotWater.Enums.EMEADomesticHotWaterState'.
    #         Path 'state', line 1, position 29.
    #     """
    # }, {
    #     "code": "InvalidInput", "message": """
    #         Error converting value 'Off'
    #         to type 'DomesticHotWater.Enums.EMEADomesticHotWaterSetpointMode'.
    #         Path 'mode', line 1, position 44.
    #     """
    # }]  # NOTE: message has been slightly edited for readability

    #
    # PART 4B: Try 'bad' task_id values...
    url_tsk = "commTasks?commTaskId=ABC"
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.GET, url_tsk, status=HTTPStatus.BAD_REQUEST
    )  # [{"code": "InvalidInput", "message": "Invalid Input."}]

    url_tsk = "commTasks?commTaskId=12345678"
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.GET, url_tsk, status=HTTPStatus.NOT_FOUND
    )  # [{"code": "CommTaskNotFound", "message": "Communication task not found."}]


# TODO: a short test
async def _test_task_id_zone(evo: EvohomeClientv2) -> None:
    """Test the task_id returned when using the vendor's RESTful APIs.

    This test can be used to prove that JSON keys are can be camelCase or PascalCase.
    """

    await evo.update(dont_update_status=True)

    for loc in evo.locations:
        for gwy in loc.gateways:
            for tcs in gwy.systems:
                if not tcs.zones:
                    continue
                zone = tcs.zones[0]
                break

    if zone is None:
        pytest.skip("No available Zone found")

    GET_URL = f"{zone._TYPE}/{zone.id}/status"
    # T_URL = f"{zone._TYPE}/{zone.id}/mode"

    #
    # PART 0: Get the initial mode...
    old_status = await should_work_v2(evo.auth, HTTPMethod.GET, GET_URL)
    assert isinstance(old_status, dict)  # mypy
    # {
    #     'zoneId': '3432576',
    #     'name': 'Main Room'
    #     'temperatureStatus': {'temperature': 25.5, 'isAvailable': True}
    #     'setpointStatus': {
    #         'targetHeatTemperature': 10.0,
    #         'setpointMode': 'FollowSchedule'
    #      }
    #     'activeFaults': []
    # }  # HTTP 200


@skipif_auth_failed
async def test_task_id_dhw(evohome_v2: EvohomeClientv2) -> None:
    """Test /commTasks?commTaskId={task_id}"""

    if not _DBG_USE_REAL_AIOHTTP:
        pytest.skip("Test is only valid with a real server")

    try:
        await _test_task_id_dhw(evohome_v2)

    except evo2.AuthenticationFailedError:
        if not _DBG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")


@skipif_auth_failed
async def _out_test_task_id_zone(evohome_v2: EvohomeClientv2) -> None:
    """Test /commTasks?commTaskId={task_id}"""

    if not _DBG_USE_REAL_AIOHTTP:
        pytest.skip("Test is only valid with a real server")

    try:
        await _test_task_id_zone(evohome_v2)

    except evo2.AuthenticationFailedError:
        if not _DBG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Unable to authenticate")



"""evohomeasync provides an async client for the Resideo TCC API."""

from __future__ import annotations

import re
from datetime import datetime as dt
from typing import TYPE_CHECKING, Any, overload

from .const import _DBG_DONT_REDACT_SECRETS, REGEX_EMAIL_ADDRESS

if TYPE_CHECKING:
    from collections.abc import Callable
    from datetime import tzinfo


REGEX_DATETIME = r"^\d{4}-\d{2}-\d{2}[T ]\d{2}:\d{2}:\d{2}"

_REDACTED_EMAIL_ADDRESS = "no-reply@redacted.xxx"
_REDACTED_STRING = "********"

_REDACTED_KEYS = (  # also keys with 'name' in them
    "city",
    "crc",
    "mac",
    "macid",
    "postcode",
    "securityquestion1",
    "securityquestion2",
    "securityquestion3",
    "streetaddress",
    "telephone",
    "zipcode",
)


def _convert_keys[T](data: T, fnc: Callable[[str], str]) -> T:
    """Recursively convert all dict keys as per some function.

    For example, converts all keys to snake_case, or CamelCase, etc.
    Used after retrieving (or before sending) JSON via the vendor API.
    """

    def recurse(data_: Any) -> Any:
        if isinstance(data_, list):
            return [recurse(i) for i in data_]

        if not isinstance(data_, dict):
            return data_

        return {fnc(k): recurse(v) for k, v in data_.items()}

    return recurse(data)  # type:ignore[no-any-return]


def _convert_vals[T](data: T, fnc: Callable[[str], str]) -> T:
    """Recursively convert all string values as per some function.

    For example, converts all isoformat string values to TZ-aware format.
    Used after retrieving (or before sending) JSON via the vendor API.
    """

    def recurse(data_: Any) -> Any:
        if isinstance(data_, dict):
            return {fnc(k): recurse(v) for k, v in data_.items()}

        if isinstance(data_, list):
            return [recurse(i) for i in data_]

        if isinstance(data_, str):
            return fnc(data_)

        return data_

    return recurse(data)  # type:ignore[no-any-return]


def as_local_time(dtm: dt | str, tzinfo: tzinfo) -> dt:
    """Convert a datetime into a aware datetime in the given TZ.

    If the datetime is naive, assume it is in the same timezone as tzinfo.
    """

    if isinstance(dtm, str):
        dtm = dt.fromisoformat(dtm)

    return dtm.replace(tzinfo=tzinfo) if dtm.tzinfo is None else dtm.astimezone(tzinfo)


_STEP_1 = re.compile(r"(.)([A-Z][a-z]+)")
_STEP_2 = re.compile(r"([a-z0-9])([A-Z])")


def camel_to_pascal(s: str) -> str:
    """Return a string convert (from camelCase) to PascalCase."""
    if " " in s:
        raise ValueError("Input string should not contain spaces")
    return s[:1].upper() + s[1:]


def camel_to_snake(s: str) -> str:
    """Return a string converted (from camelCase) to snake_case."""
    if " " in s:
        raise ValueError("Input string should not contain spaces")
    return _STEP_2.sub(r"\1_\2", _STEP_1.sub(r"\1_\2", s)).lower()


def snake_to_camel(s: str) -> str:
    """Return a string converted (from snake_case) to camelCase."""
    if " " in s:
        raise ValueError("Input string should not contain spaces")
    components = s.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


def noop[T](s: T) -> T:
    """Return a value (usually a string) unconverted."""
    return s


def convert_keys_to_camel_case[T](data: T) -> T:
    """Recursively convert all dict keys from snake_case to camelCase.

    Used before sending JSON to the vendor API.
    """
    return _convert_keys(data, snake_to_camel)


def convert_keys_to_snake_case[T](data: T) -> T:
    """Recursively convert all dict keys from camelCase to snake_case.

    Used after retrieving JSON from the vendor API.
    """
    return _convert_keys(data, camel_to_snake)


def convert_naive_dtm_strs_to_aware[T](data: T, tzinfo: tzinfo) -> T:
    """Recursively convert TZ-naive datetime strings to TZ-aware.

    Does not convert TZ-aware strings, even if they're from a different TZ.
    Used after retrieving JSON from the vendor API.
    """

    def recurse(data_: Any) -> Any:  # noqa: PLR0911
        if isinstance(data_, dict):
            return {k: recurse(v) for k, v in data_.items()}

        if isinstance(data_, list):
            return [recurse(i) for i in data_]

        if not isinstance(data_, str) or not re.match(REGEX_DATETIME, data_):
            return data_

        try:
            d = dt.fromisoformat(data_)
        except ValueError:
            return data_

        if d.tzinfo is None:  # e.g. 2023-11-30T22:10:00
            return d.replace(tzinfo=tzinfo).isoformat()

        if data_.endswith("Z"):  # e.g. 2023-11-30T22:10:00Z
            return d.astimezone(tzinfo).isoformat()

        return data_  # e.g. 2023-11-30T22:10:00+00:00

    return recurse(data)  # type:ignore[no-any-return]


@overload
def redact(value: bool) -> bool | None: ...  # type: ignore[overload-overlap] # noqa: FBT001
@overload
def redact(value: int) -> int: ...
@overload
def redact(value: str) -> str: ...


def redact(value: bool | int | str) -> bool | int | str | None:  # noqa: FBT001
    """Redact a value (usually to protect secrets when logging)."""

    if _DBG_DONT_REDACT_SECRETS:
        return value
    if isinstance(value, bool):
        return None
    if isinstance(value, int):
        return 0
    if not isinstance(value, str):
        raise TypeError(f"redact() expects bool | int | str, got {type(value)}")
    if REGEX_EMAIL_ADDRESS.match(value):
        return _REDACTED_EMAIL_ADDRESS
    return _REDACTED_STRING


def redact_secrets[T](data: T) -> T:
    """Recursively redact all dict/list values that might be secrets.

    Used when logging JSON received from the vendor API.
    """

    def _redact(key: str, val: Any) -> Any:
        if not isinstance(val, str):
            return redact(val)
        if REGEX_EMAIL_ADDRESS.match(val):
            return _REDACTED_EMAIL_ADDRESS
        if "name" in key.lower():
            return val[:2].ljust(len(val), "*")
        return "".join("*" if char != " " else " " for char in val)

    def should_redact(key: Any) -> bool:
        # unfortunately, also redacts 'displayName' (is under 'timeZone')
        return isinstance(key, str) and (
            "name" in key.lower() or key.lower() in _REDACTED_KEYS
        )

    def recurse(data_: Any) -> Any:
        if isinstance(data_, list):  # or Sequence?
            return [recurse(i) for i in data_]

        if isinstance(data_, tuple):
            return tuple(recurse(i) for i in data_)

        if not isinstance(data_, dict):  # Mapping?
            return data_

        return {
            k: _redact(k, v) if should_redact(k) else recurse(v)
            for k, v in data_.items()
        }

    return data if _DBG_DONT_REDACT_SECRETS else recurse(data)



"""Validate the handling of the vendor's v2 APIs (URLs) for Authentication.

This is used to:
a) document the RESTful API that is provided by the vendor
b) confirm the faked server (if any) is behaving as per a)

Testing is at HTTP request layer (e.g. POST).
Everything to/from the RESTful API is in camelCase (so those schemas are used).
"""

from __future__ import annotations

import random
import socket
import string
from http import HTTPStatus
from typing import TYPE_CHECKING
from urllib.parse import urlparse

import aiohttp
import pytest

from evohomeasync2.schemas import (
    TCC_ERROR_RESPONSE,
    TCC_GET_USR_ACCOUNT,
    TCC_POST_OAUTH_TOKEN,
    TCC_STATUS_RESPONSE,
)
from tests.const import (
    _DBG_TEST_CRED_URLS,
    _DBG_USE_REAL_AIOHTTP,
    HEADERS_BASE,
    HEADERS_CRED_V2,
    URL_BASE_V2,
    URL_CRED_V2,
)

if TYPE_CHECKING:
    from evohomeasync2.schemas import (
        TccErrorResponseT,
        TccFailureResponseT,
        TccOAuthTokenResponseT,
        TccUsrAccountResponseT,
    )


async def _skip_if_too_many_requests(rsp: aiohttp.ClientResponse) -> None:
    if rsp.status != HTTPStatus.TOO_MANY_REQUESTS:  # 429
        return

    response: TccErrorResponseT = await rsp.json()

    # the expected response for TOO_MANY_REQUESTS
    """
        {"error": "attempt_limit_exceeded"}
    """

    assert response["error"] == "attempt_limit_exceeded"
    TCC_ERROR_RESPONSE(response)

    pytest.skip("Too many requests")


@pytest.mark.skipif(not _DBG_USE_REAL_AIOHTTP, reason="requires vendor's webserver")
async def _test_bad0(  # invalid server certificate
    client_session: aiohttp.ClientSession,
) -> None:
    """Test authentication flow with an invalid server certificate."""

    #
    # TEST 0: invalid server certificate
    parsed_url = urlparse(URL_CRED_V2)
    ip_address = socket.gethostbyname(parsed_url.hostname)  # type: ignore[arg-type]
    url = f"{parsed_url.scheme}://{ip_address}{parsed_url.path}"  # path[:1] == "/"

    # aiohttp.client_exceptions.ClientConnectorCertificateError:
    # Cannot connect to host 135.224.164.118:443 ssl:True
    # [SSLCertVerificationError: ... certificate is not valid for ...]

    with pytest.raises(aiohttp.ClientConnectorCertificateError):
        await client_session.post(url, headers=HEADERS_CRED_V2, data={})


@pytest.mark.skipif(not _DBG_USE_REAL_AIOHTTP, reason="requires vendor's webserver")
async def test_bad1(  # bad credentials (client_id/secret)
    client_session: aiohttp.ClientSession,
) -> None:
    """Test authentication flow with bad credentials (client_id/secret)."""

    #
    # TEST 1: bad credentials (client_id/secret) -> HTTPStatus.BAD_REQUEST
    data = {
        "grant_type": "password",
        "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",  # EMEA-V1-Get-Current-User-Account",
        "Username": random.choice(string.ascii_letters),  # noqa: S311
        "Password": "",
    }

    async with client_session.post(
        URL_CRED_V2, headers=HEADERS_CRED_V2, data=data
    ) as rsp:
        await _skip_if_too_many_requests(rsp)
        assert rsp.status == HTTPStatus.BAD_REQUEST  # 400

        response: TccErrorResponseT = await rsp.json()

        # the expected response for bad credentials
        """
            {"error": "invalid_grant"}
        """

    assert response["error"] == "invalid_grant"
    TCC_ERROR_RESPONSE(response)


@pytest.mark.skipif(not _DBG_USE_REAL_AIOHTTP, reason="requires vendor's webserver")
async def test_bad2(  # bad access token
    client_session: aiohttp.ClientSession,
) -> None:
    """Test authentication flow with an invalid/expired access token."""

    # pre-requisite data
    access_token = "bad access token "  # noqa: S105

    #
    # TEST 9: bad access token -> HTTPStatus.UNAUTHORIZED
    url = f"{URL_BASE_V2}/userAccount"
    headers = HEADERS_BASE | {"Authorization": "Bearer " + access_token}

    async with client_session.get(url, headers=headers) as rsp:
        await _skip_if_too_many_requests(rsp)
        assert rsp.status == HTTPStatus.UNAUTHORIZED  # 401

        response: list[TccFailureResponseT] = await rsp.json()

        # the expected response for bad access tokens
        """
            [{
                "code": "Unauthorized",
                "message": "Unauthorized."
            }]
        """

    assert isinstance(response, list) and response[0]["code"] == "Unauthorized"  # noqa: PT018
    TCC_STATUS_RESPONSE(response)


@pytest.mark.skipif(not _DBG_USE_REAL_AIOHTTP, reason="requires vendor's webserver")
async def test_bad3(  # bad credentials (refresh token)
    client_session: aiohttp.ClientSession,
) -> None:
    """Test authentication flow with invalid/unknown credentials (refresh token)."""

    # pre-requisite data
    refresh_token = "invalid refresh token "  # noqa: S105

    #
    # TEST 2: bad credentials (refresh token) -> HTTPStatus.BAD_REQUEST
    data = {
        "grant_type": "refresh_token",
        "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",
        "refresh_token": refresh_token,
    }

    async with client_session.get(
        URL_CRED_V2, headers=HEADERS_CRED_V2, data=data
    ) as rsp:
        await _skip_if_too_many_requests(rsp)
        assert rsp.status == HTTPStatus.BAD_REQUEST  # 400

        response: TccErrorResponseT = await rsp.json()

        # the expected response for bad refresh tokens
        """
            {"error": "invalid_grant"}
        """

    assert response["error"] == "invalid_grant"
    TCC_ERROR_RESPONSE(response)


@pytest.mark.skipif(not _DBG_USE_REAL_AIOHTTP, reason="requires vendor's webserver")
@pytest.mark.skipif(not _DBG_TEST_CRED_URLS, reason="may invalidate credentials cache")
async def test_good(  # good credentials
    client_session: aiohttp.ClientSession,
    credentials: tuple[str, str],
) -> None:
    """Test authentication flow (and authorization) with good credentials."""

    #
    # TEST 1: good credentials (client_id/secret) -> HTTPStatus.OK
    data = {
        "grant_type": "password",
        "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",  # EMEA-V1-Get-Current-User-Account",
        "Username": credentials[0],
        "Password": credentials[1],
    }

    async with client_session.post(
        URL_CRED_V2, headers=HEADERS_CRED_V2, data=data
    ) as rsp:
        await _skip_if_too_many_requests(rsp)
        assert rsp.status == HTTPStatus.OK  # 200 (400 is bad credentials)

        user_auth: TccOAuthTokenResponseT = await rsp.json()

        # the expected response for good credentials
        """
            {
                "access_token": "HojUMRvmn...",
                "token_type": "bearer",
                "expires_in": 1799,
                "refresh_token": "vdBdHjxK...",
                # "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",  # optional
            }
        """

    #
    TCC_POST_OAUTH_TOKEN(user_auth)

    # #################################################################################

    # Check the access token by accessing a resource...
    access_token = user_auth["access_token"]

    #
    # TEST 9: good access token -> HTTPStatus.OK  # 200 (401 is bad token)
    url = f"{URL_BASE_V2}/userAccount"
    headers = HEADERS_BASE | {"Authorization": "Bearer " + access_token}

    async with client_session.get(url, headers=headers) as rsp:
        await _skip_if_too_many_requests(rsp)
        assert rsp.status == HTTPStatus.OK  # 200

        user_info: TccUsrAccountResponseT = await rsp.json()

        # the expected response for good access token
        """
            {
                "userId": "1234567",
                "username": "username@email.com",
                "firstname": "David",
                "lastname": "Smith",
                "streetAddress": "1 Main Street",
                "city": "London",
                "postcode": "E1 1AA",
                "country": "UnitedKingdom",
                "language": "enGB"
            }
        """

    assert user_info["username"] == credentials[0]
    TCC_GET_USR_ACCOUNT(user_info)

    # #################################################################################

    # Check the refresh token by requesting another access token...
    refresh_token = user_auth["refresh_token"]

    #
    # TEST 2: good credentials (refresh token)-> HTTPStatus.OK
    data = {
        "grant_type": "refresh_token",
        "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",
        "refresh_token": refresh_token,
    }

    async with client_session.post(
        URL_CRED_V2, headers=HEADERS_CRED_V2, data=data
    ) as rsp:
        await _skip_if_too_many_requests(rsp)
        assert rsp.status == HTTPStatus.OK  # 200

        user_auth = await rsp.json()  # TccOAuthTokenResponseT

        # the expected response for good refresh tokens
        """
            {
                "access_token": "HojUMRvmn...",
                "token_type": "bearer",
                "expires_in": 1799,
                "refresh_token": "vdBdHjxK...",
                # "scope": "EMEA-V1-Basic EMEA-V1-Anonymous",  # optional
            }
        """

    #
    TCC_POST_OAUTH_TOKEN(user_auth)



"""Provides handling of TCC DHW zones."""

# TODO: extend set_mode() for non-evohome modes

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Final

from _evohome.helpers import as_local_time, camel_to_snake

from . import exceptions as exc
from .const import (
    API_STRFTIME,
    SZ_ALLOWED_MODES,
    SZ_DHW_ID,
    SZ_DHW_STATE_CAPABILITIES_RESPONSE,
    SZ_MODE,
    SZ_SCHEDULE_CAPABILITIES_RESPONSE,
    SZ_STATE,
    SZ_STATE_STATUS,
)
from .schemas import (
    DhwState,
    TccSetDhwModeT,
    ZoneMode,
    factory_dhw_schedule,
    factory_dhw_status,
)
from .schemas.const import S2_MODE, S2_STATE, S2_UNTIL_TIME, EntityType
from .zone import _ZoneBase

if TYPE_CHECKING:
    from datetime import datetime as dt

    import voluptuous as vol

    from . import ControlSystem
    from .schemas.typedefs import (
        DayOfWeekDhwT,
        EvoDhwConfigEntryT,
        EvoDhwConfigResponseT,
        EvoDhwScheduleCapabilitiesResponseT,
        EvoDhwStateCapabilitiesResponseT,
        EvoDhwStateStatusResponseT,
        EvoDhwStatusResponseT,
    )


class HotWater(_ZoneBase):
    """Instance of a TCS's DHW zone (domesticHotWater)."""

    _TYPE = EntityType.DHW

    SCH_SCHEDULE: vol.Schema = factory_dhw_schedule(camel_to_snake)
    SCH_STATUS: vol.Schema = factory_dhw_status(camel_to_snake)

    def __init__(self, tcs: ControlSystem, config: EvoDhwConfigResponseT) -> None:
        super().__init__(config[SZ_DHW_ID], tcs)

        self._config: Final[EvoDhwConfigEntryT] = config  # type: ignore[misc]
        self._status: EvoDhwStatusResponseT | None = None

        self._schedule: list[DayOfWeekDhwT] | None = None  # type: ignore[assignment]

    @property
    def config(self) -> EvoDhwConfigEntryT:
        """Return the latest config of the entity."""
        return self._config

    @property
    def status(self) -> EvoDhwStatusResponseT:
        """Return the latest status of the entity."""
        return super().status  # type: ignore[return-value]

    # Config attrs...

    @cached_property
    def name(self) -> str:
        return "Domestic Hot Water"

    @cached_property
    def type(self) -> str:
        return "DomesticHotWater"

    @cached_property  # NOTE: renamed config key: was schedule_capabilities_response
    def schedule_capabilities(self) -> EvoDhwScheduleCapabilitiesResponseT:
        """
        "scheduleCapabilitiesResponse": {
          "maxSwitchpointsPerDay": 6,
          "minSwitchpointsPerDay": 1,
          "timingResolution": "00:10:00"
        }
        """

        return self._config[SZ_SCHEDULE_CAPABILITIES_RESPONSE]

    @cached_property  # NOTE: renamed config key: was dhw_state_capabilities_response
    def state_capabilities(self) -> EvoDhwStateCapabilitiesResponseT:
        """
        "dhwStateCapabilitiesResponse": {
            "allowedStates": ["On", "Off"],
            "allowedModes": ["FollowSchedule", "PermanentOverride", "TemporaryOverride"],
            "maxDuration": "1.00:00:00",
            "timingResolution": "00:10:00"
        },
        """

        return self._config[SZ_DHW_STATE_CAPABILITIES_RESPONSE]

    @cached_property
    def allowed_modes(self) -> tuple[ZoneMode, ...]:
        return tuple(self.state_capabilities[SZ_ALLOWED_MODES])

    @cached_property
    def allowed_states(self) -> tuple[DhwState, ...]:
        return tuple(self.state_capabilities["allowed_states"])

    # Status (state) attrs & methods...

    @property
    def state_status(self) -> EvoDhwStateStatusResponseT:
        """
        "stateStatus": {"state": "Off", "mode": "PermanentOverride"}
        "stateStatus": {
            "state": "Off",
            "mode": "TemporaryOverride",
            until": "2023-11-30T22:10:00Z"
        }
        """

        if self._status is None:
            raise exc.InvalidStatusError(f"{self} has no state, has it been fetched?")
        return self._status[SZ_STATE_STATUS]

    @property
    def mode(self) -> ZoneMode:
        return self.state_status[SZ_MODE]

    @property
    def state(self) -> DhwState:
        return self.state_status[SZ_STATE]

    @property
    def until(self) -> dt | None:
        if (until := self.state_status.get("until")) is None:
            return None
        return as_local_time(until, self.location.tzinfo)

    async def _set_mode(self, mode: TccSetDhwModeT) -> None:
        """Set the DHW mode (state)."""

        # Issue a warning if we fail some basic sanity checks...
        if mode[S2_MODE] not in self.allowed_modes:
            self._logger.warning(f"{self}: Unsupported/unknown {S2_MODE}: {mode}")

        if not (state := mode.get(S2_STATE)):
            if mode[S2_MODE] != ZoneMode.FOLLOW_SCHEDULE:
                self._logger.warning(f"{self}: Missing/invalid {S2_STATE}: {mode}")

        elif state not in self.allowed_states:
            self._logger.warning(f"{self}: Unsupported/unknown {S2_STATE}: {mode}")

        # Call the API...
        await self._auth.put(f"{self._TYPE}/{self.id}/state", json=dict(mode))

    async def reset(self) -> None:
        """Cancel any override and allow the DHW to follow its schedule."""
        await self._set_mode({S2_MODE: ZoneMode.FOLLOW_SCHEDULE})

    async def set_mode(self, state: DhwState, /, *, until: dt | None = None) -> None:
        """Set the DHW mode (state)."""

        mode: TccSetDhwModeT

        if until is None:
            mode = {
                S2_MODE: ZoneMode.PERMANENT_OVERRIDE,
                S2_STATE: state,
            }
        else:
            mode = {  # NOTE: UNTIL_TIME, not TIME_UNTIL
                S2_MODE: ZoneMode.TEMPORARY_OVERRIDE,
                S2_STATE: state,
                S2_UNTIL_TIME: until.strftime(API_STRFTIME),
            }

        await self._set_mode(mode)

    async def off(self, /, *, until: dt | None = None) -> None:
        """Set the DHW off until a given time, or permanently."""
        await self.set_mode(DhwState.OFF, until=until)

    async def on(self, /, *, until: dt | None = None) -> None:
        """Set the DHW on until a given time, or permanently."""
        await self.set_mode(DhwState.ON, until=until)

    # NOTE: this wrapper exists only for typing purposes
    async def get_schedule(self) -> list[DayOfWeekDhwT]:  # type: ignore[override]
        """Get the schedule for this DHW zone."""
        return await super().get_schedule()  # type: ignore[return-value]

    # NOTE: this wrapper exists only for typing purposes
    async def set_schedule(self, schedule: list[DayOfWeekDhwT] | str) -> None:  # type: ignore[override]
        """Set the schedule for this DHW zone."""
        await super().set_schedule(schedule)  # type: ignore[arg-type]



"""Tests for evohome-async."""

from __future__ import annotations

from typing import Final

from _evohome.auth import HOSTNAME
from evohomeasync2.auth import _APPLICATION_ID as _APPLICATION_ID_V2

#
# all _DBG_* flags are only for dev/test and should be False for published code
_DBG_DISABLE_STRICT_ASSERTS = False  # of response content-type, schema
_DBG_TEST_CRED_URLS = False  # avoid 429s: dont invalidate the credential cache
_DBG_USE_REAL_AIOHTTP = False  # use 'real' aiohttp to reach vendor's servers

#
# used to construct the default token cache
TEST_USERNAME: Final = "username@email.com"
TEST_PASSWORD: Final = "P@ssw0rd!!"  # noqa: S105

# vendors API URLs - the older API
URL_CRED_V0 = f"https://{HOSTNAME}/WebAPI/api/session"
URL_BASE_V0 = f"https://{HOSTNAME}/WebAPI/api"

# - the newer API
URL_CRED_V2 = f"https://{HOSTNAME}/Auth/OAuth/Token"
URL_BASE_V2 = f"https://{HOSTNAME}/WebAPI/emea/api/v1"

HEADERS_BASE = {
    "Accept": "application/json",
    "Connection": "Keep-Alive",
}
HEADERS_CRED_V0 = HEADERS_BASE | {
    "Cache-Control": "no-cache, no-store",
    "Pragma": "no-cache",
}
HEADERS_CRED_V2 = HEADERS_BASE | {
    "Cache-Control": "no-cache, no-store",
    "Pragma": "no-cache",
    "Authorization": "Basic " + _APPLICATION_ID_V2,
}



"""evohome-async - validate the v2 API for DHW."""

from __future__ import annotations

from datetime import UTC, datetime as dt, timedelta as td
from http import HTTPMethod
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import AsyncMock, patch

import pytest

from evohomeasync2.schemas.const import DhwState

from .conftest import FIXTURES_V2

if TYPE_CHECKING:
    from freezegun.api import FrozenDateTimeFactory

    from evohomeasync2 import EvohomeClient


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    folders = [
        p for p in Path(FIXTURES_V2).glob("*") if p.is_dir() and p.name == "default"
    ]
    metafunc.parametrize(
        "fixture_folder", sorted(folders), ids=(p.name for p in sorted(folders))
    )


# Test the ControlSystem APIs (incomplete)...
# NOTE: not all systems support all modes, below we only test evohome modes


async def test_ctl_mode_reset(  # TODO: test systems without AutoWithReset
    evohome_v2: EvohomeClient,
) -> None:
    """Test ControlSystem.reset() method."""

    tcs = evohome_v2.tcs

    url = f"temperatureControlSystem/{tcs.id}/mode"
    mode = {
        "systemMode": "AutoWithReset",  # SystemMode.AUTO_WITH_RESET,
        "permanent": True,
    }

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await tcs.reset()

    mock_put.assert_awaited_once_with(HTTPMethod.PUT, url, json=mode)


CTL_APIS_SANS_UNTIL = {  # system mode APIs that can not take an until kwarg
    "set_auto": "Auto",  # SystemMode.AUTO,
    "set_heatingoff": "HeatingOff",  # SystemMode.HEATING_OFF,
}


@pytest.mark.parametrize("api_name", CTL_APIS_SANS_UNTIL)
async def test_ctl_modes_sans_until(
    evohome_v2: EvohomeClient,
    api_name: str,
) -> None:
    """Test ControlSystem.set_auto() method."""

    tcs = evohome_v2.tcs

    url = f"temperatureControlSystem/{tcs.id}/mode"
    mode = {
        "systemMode": CTL_APIS_SANS_UNTIL[api_name],
        "permanent": True,
    }

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await getattr(tcs, api_name)()

    mock_put.assert_awaited_once_with(HTTPMethod.PUT, url, json=mode)

    with (
        patch("_evohome.auth.AbstractAuth.request", new_callable=AsyncMock) as mock_put,
        pytest.raises(TypeError),  # got an unexpected keyword argument 'until'
    ):
        await getattr(tcs, api_name)(until=dt.now(tz=UTC) + td(days=3))


CTL_APIS_WITH_UNTIL = {  # system mode APIs that can take an until kwarg
    "set_away": "Away",  # SystemMode.AWAY,
    "set_custom": "Custom",  # SystemMode.CUSTOM,
    "set_dayoff": "DayOff",  # SystemMode.DAY_OFF,
    "set_eco": "AutoWithEco",  # SystemMode.AUTO_WITH_ECO,
}


@pytest.mark.parametrize("api_name", CTL_APIS_WITH_UNTIL)
async def test_ctl_modes_with_until(
    evohome_v2: EvohomeClient,
    api_name: str,
    freezer: FrozenDateTimeFactory,
) -> None:
    """Test ControlSystem.set_*() methods (that can take an until kwarg)."""

    tcs = evohome_v2.tcs

    url = f"temperatureControlSystem/{tcs.id}/mode"
    mode = {
        "systemMode": CTL_APIS_WITH_UNTIL[api_name],
        "permanent": True,
    }

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await getattr(tcs, api_name)()

    mock_put.assert_awaited_once_with(HTTPMethod.PUT, url, json=mode)

    freezer.move_to("2025-07-10T12:00:00Z")

    mode = {
        "systemMode": CTL_APIS_WITH_UNTIL[api_name],
        "permanent": False,
        "timeUntil": "2025-07-13T12:00:00Z",
    }

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await getattr(tcs, api_name)(until=dt.now(tz=UTC) + td(days=3))

    mock_put.assert_awaited_once_with(HTTPMethod.PUT, url, json=mode)


# Test the HotWater APIs...


async def test_dhw_mode_off(
    evohome_v2: EvohomeClient,
) -> None:
    """Test HotWater.off() method."""

    dhw = evohome_v2.tcs.hotwater
    assert dhw is not None

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await dhw.off()

    EXPECTED_JSON = {
        "mode": "PermanentOverride",  # ZoneMode.PERMANENT_OVERRIDE,
        "state": "Off",  # #            DhwState.OFF,
    }

    mock_put.assert_awaited_once()

    assert mock_put.call_args[0][0] == HTTPMethod.PUT
    assert mock_put.call_args[0][1] == f"domesticHotWater/{dhw.id}/state"
    assert mock_put.call_args[1] == {"json": EXPECTED_JSON}


async def test_dhw_mode_on(
    evohome_v2: EvohomeClient,
) -> None:
    """Test HotWater.on() method."""

    dhw = evohome_v2.tcs.hotwater
    assert dhw is not None

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await dhw.on()

    mock_put.assert_awaited_once()

    EXPECTED_JSON = {
        "mode": "PermanentOverride",  # ZoneMode.PERMANENT_OVERRIDE,
        "state": "On",  # #             DhwState.ON,
    }

    assert mock_put.call_args[0][0] == HTTPMethod.PUT
    assert mock_put.call_args[0][1] == f"domesticHotWater/{dhw.id}/state"
    assert mock_put.call_args[1] == {"json": EXPECTED_JSON}


async def test_dhw_mode_reset(
    evohome_v2: EvohomeClient,
) -> None:
    """Test HotWater.reset() method."""

    dhw = evohome_v2.tcs.hotwater
    assert dhw is not None

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await dhw.reset()

    mock_put.assert_awaited_once()

    EXPECTED_JSON = {
        "mode": "FollowSchedule",  # ZoneMode.FOLLOW_SCHEDULE,
    }

    assert mock_put.call_args[0][0] == HTTPMethod.PUT
    assert mock_put.call_args[0][1] == f"domesticHotWater/{dhw.id}/state"
    assert mock_put.call_args[1] == {"json": EXPECTED_JSON}


async def test_dhw_set_mode(
    evohome_v2: EvohomeClient,
    freezer: FrozenDateTimeFactory,
) -> None:
    """Test HotWater.on() method."""

    dhw = evohome_v2.tcs.hotwater
    assert dhw is not None

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await dhw.set_mode(DhwState.OFF)

    mock_put.assert_awaited_once()

    EXPECTED_JSON = {
        "mode": "PermanentOverride",  # ZoneMode.PERMANENT_OVERRIDE,
        "state": "Off",  # #            DhwState.OFF,
    }

    assert mock_put.call_args[0][0] == HTTPMethod.PUT
    assert mock_put.call_args[0][1] == f"domesticHotWater/{dhw.id}/state"
    assert mock_put.call_args[1] == {"json": EXPECTED_JSON}

    freezer.move_to("2025-07-10T12:00:00Z")

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await dhw.set_mode(DhwState.ON, until=dt.now(tz=UTC) + td(hours=3))

    mock_put.assert_awaited_once()

    EXPECTED_JSON = {
        "mode": "TemporaryOverride",  # ZoneMode.TEMPORARY_OVERRIDE,
        "state": "On",  # #             DhwState.ON,
        "untilTime": "2025-07-10T15:00:00Z",
    }

    assert mock_put.call_args[0][0] == HTTPMethod.PUT
    assert mock_put.call_args[0][1] == f"domesticHotWater/{dhw.id}/state"
    assert mock_put.call_args[1] == {"json": EXPECTED_JSON}


# Test the Zone APIs...


async def test_zon_mode_reset(
    evohome_v2: EvohomeClient,
) -> None:
    """Test Zone.reset() method."""

    zone = evohome_v2.tcs.zones[0]

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await zone.reset()

    mock_put.assert_awaited_once()

    EXPECTED_JSON = {
        "setpointMode": "FollowSchedule",  # ZoneMode.FOLLOW_SCHEDULE,
    }

    assert mock_put.call_args[0][0] == HTTPMethod.PUT
    assert mock_put.call_args[0][1] == f"temperatureZone/{zone.id}/heatSetpoint"
    assert mock_put.call_args[1] == {"json": EXPECTED_JSON}


async def test_zon_mode_set_temperature(
    evohome_v2: EvohomeClient,
    freezer: FrozenDateTimeFactory,
) -> None:
    """Test Zone.set_temperature() method."""

    zone = evohome_v2.tcs.zones[0]

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await zone.set_temperature(19.5)

    mock_put.assert_awaited_once()

    EXPECTED_JSON = {
        "setpointMode": "PermanentOverride",  # ZoneMode.PERMANENT_OVERRIDE,
        "heatSetpointValue": 19.5,
    }

    assert mock_put.call_args[0][0] == HTTPMethod.PUT
    assert mock_put.call_args[0][1] == f"temperatureZone/{zone.id}/heatSetpoint"
    assert mock_put.call_args[1] == {"json": EXPECTED_JSON}

    freezer.move_to("2025-07-10T12:00:00Z")

    with patch(
        "_evohome.auth.AbstractAuth.request", new_callable=AsyncMock
    ) as mock_put:
        await zone.set_temperature(20.5, until=dt.now(tz=UTC) + td(hours=1))

    mock_put.assert_awaited_once()

    EXPECTED_JSON = {
        "setpointMode": "TemporaryOverride",  # ZoneMode.TEMPORARY_OVERRIDE,
        "heatSetpointValue": 20.5,
        "timeUntil": "2025-07-10T13:00:00Z",
    }

    assert mock_put.call_args[0][0] == HTTPMethod.PUT
    assert mock_put.call_args[0][1] == f"temperatureZone/{zone.id}/heatSetpoint"
    assert mock_put.call_args[1] == {"json": EXPECTED_JSON}



"""Validate the instantiation/update of each client.

Performs the bare minimum of tests.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from _evohome import exceptions as exc
from tests.const import _DBG_USE_REAL_AIOHTTP

from .common import skipif_auth_failed

if TYPE_CHECKING:
    from tests.conftest import EvohomeClientv0, EvohomeClientv2


#######################################################################################


@skipif_auth_failed
@pytest.mark.skipif(not _DBG_USE_REAL_AIOHTTP, reason="requires vendor's webserver")
async def test_update_v0(evohome_v0: EvohomeClientv0) -> None:
    """Make a minimal test of instantiation/update of the v0 client."""

    with pytest.raises(exc.InvalidConfigError):
        assert evohome_v0.user_account

    await evohome_v0.update()

    assert evohome_v0.user_account


@skipif_auth_failed
@pytest.mark.skipif(not _DBG_USE_REAL_AIOHTTP, reason="requires vendor's webserver")
async def test_update_v2(evohome_v2: EvohomeClientv2) -> None:
    """Make a minimal test of instantiation/update of the v2 client."""

    with pytest.raises(exc.InvalidConfigError):
        assert evohome_v2.user_account

    await evohome_v2.update(dont_update_status=True)

    assert evohome_v2.user_account
